from __future__ import annotations

import re
from abc import ABC, abstractmethod
from enum import IntEnum, StrEnum
from ipaddress import IPv4Address, IPv6Address
from typing import TYPE_CHECKING, Any, TypeAlias, Union, overload

if TYPE_CHECKING:
    from collections.abc import Set as AbstractSet


from ..exceptions import CommandError

#: List of characters in a string that need an escape.
_AUTO_CHARS: AbstractSet[str] = frozenset({' ', '"', '\\'})

#: All types allowed as a key in a keyword argument.
KeyTypes: TypeAlias = Union[str, None]  # noqa: UP007

#: All types allowed as a value, either for a keyword or a simple string.
ValueTypes: TypeAlias = Union[  # noqa: UP007
    IPv4Address,
    IPv6Address,
    IntEnum,
    StrEnum,
    int,
    str,
]


@overload
def _serialize_value(value: None, *, allow_none: bool) -> None: ...


@overload
def _serialize_value(value: ValueTypes, *, allow_none: bool) -> str: ...


def _serialize_value(value: Any, *, allow_none: bool = False) -> str | None:
    """Serialize a single value to a string."""
    result: str | None

    match value:
        case IntEnum() | StrEnum():
            result = str(value.value)
        case None:
            if allow_none is False:
                msg = 'Value cannot be None.'
                raise CommandError(msg)
            result = None
        case IPv4Address() | IPv6Address() | int() | str():
            result = str(value)
        case _:
            msg = f'Type {type(value).__name__} cannot be serialized to a string.'
            raise CommandError(msg)

    return result


class QuoteStyle(IntEnum):
    """Set the type of quote to use."""

    #: No quote are added around the value, no checks are being performed.
    NEVER = 0

    #: No quote are added around the value, check input to ensure that.
    NEVER_ENSURE = 1

    #: Value is always enclosed with quotes.
    ALWAYS = 2

    #: Automatically determine the quoting style.
    AUTO = 3

    @staticmethod
    def should_have_quotes(text: str) -> bool:
        """
        Tell whether the provided `text` should have quotes.

        Args:
            text: Input text to check for quotes.

        Returns:
            Whether the input text should be enclosed with quotes.

        """
        return any(c in text for c in _AUTO_CHARS)

    def escape(self, text: str) -> str:
        """
        Escape the provided text, if needed.

        Args:
            text: string value to quote according to the current style

        Returns:
            The input value quoted according to the current style.

        """
        do_quote = False

        match self.value:
            case QuoteStyle.ALWAYS.value:
                do_quote = True
            case QuoteStyle.AUTO.value:
                do_quote = self.should_have_quotes(text)
            case QuoteStyle.NEVER_ENSURE.value:
                if self.should_have_quotes(text):
                    msg = 'Argument is only safe with quotes'
                    raise CommandError(msg)

        if do_quote:
            text = '"' + re.sub(r'([\\"])', r'\\\1', text) + '"'
        return text


class BaseArgument(ABC):
    """Base class for any command argument."""

    @abstractmethod
    def __str__(self) -> str:
        """Serialize the argument to string."""


class ArgumentKeyword(BaseArgument):
    """Describe a keyword argument."""

    def __init__(
        self,
        key: KeyTypes,
        value: ValueTypes | None,
        *,
        quotes: QuoteStyle = QuoteStyle.AUTO,
    ) -> None:
        """
        Create a new keyword argument.

        Important:
            Both ``key`` and ``value`` cannot be :obj:`None` at the same time.

        Note:
            When value is :obj:`None`, quotes is enforced to :data:`QuoteStyle.NEVER`.

        Args:
            key: Key part of the keyword, if any.
            value: Value part of the keyword, if any.

        Raises:
            CommandError: when ``key`` and ``value`` are both :obj:`None`.

        Keyword Args:
            quotes: Tell how to quote the value part when serialized.

        """
        # When value is None, we are treated as a flag and never need to have quotes.
        if value is None:
            quotes = QuoteStyle.NEVER
            if key is None:
                msg = 'Both key and value cannot be None.'
                raise CommandError(msg)

        self._key = _serialize_value(key, allow_none=True)
        self._value = _serialize_value(value, allow_none=True)
        self._quotes = quotes

    def __str__(self) -> str:
        """Serialize the argument to a string."""
        if self._value is None:
            # This check was already performed during __init__.
            return self._key  # type: ignore[return-value]

        value = self._quotes.escape(self._value)
        if self._key is None:
            return value

        return f'{self._key}={value}'

    @property
    def key(self) -> str | None:
        """Get the key part of the keyword argument."""
        return self._key

    @property
    def value(self) -> str | None:
        """Get the value of the keyword argument as a string."""
        return self._value

    @property
    def quotes(self) -> QuoteStyle:
        """Get the applied quoting style."""
        return self._quotes


class ArgumentString(BaseArgument):
    """Describe a string argument."""

    def __init__(self, value: ValueTypes, *, safe: bool = False) -> None:
        """
        Create a new string argument.

        Args:
            value: Raw string value.

        Keyword Args:
            safe: Whether the caller ensures that the value contains no space.

        Note:
            This value cannot contain spaces.

        """
        value_str = _serialize_value(value, allow_none=False)
        if not safe and ' ' in value_str:
            msg = f"Invalid space in argument '{value_str}'"
            raise CommandError(msg)

        self._value = value_str

    def __str__(self) -> str:
        """Serialize the argument to a string."""
        return self._value

    @property
    def value(self) -> str:
        """Get the value of the string argument."""
        return self._value



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import (
    Mapping,
    Sequence,
    Set as AbstractSet,
)
from dataclasses import dataclass, field
from enum import StrEnum
from functools import cached_property
from typing import TYPE_CHECKING, Annotated, Any, ClassVar, Literal, Self, TypeAlias, Union

from pydantic import BeforeValidator, Discriminator, Field, NonNegativeInt, Tag, TypeAdapter

from .exceptions import MessageError, ReplySyntaxError
from .reply import ReplyGetMap
from .structures import (
    CircuitBuildFlags,
    CircuitCloseReason,
    CircuitEvent,
    CircuitHiddenServicePow,
    CircuitHiddenServiceState,
    CircuitPurpose,
    CircuitStatus,
    GuardEventStatus,
    HiddenServiceAddress,
    HiddenServiceAddressV2,
    HiddenServiceAddressV3,
    HsDescAction,
    HsDescAuthTypeStr,
    HsDescFailReason,
    HsDescV2,
    HsDescV3,
    LivenessStatus,
    LogSeverity,
    LongServerName,
    OrConnCloseReason,
    OrConnStatus,
    RemapSource,
    RouterStatus,
    Signal,
    StatusActionClient,
    StatusActionGeneral,
    StatusActionServer,
    StatusClientBootstrap,
    StatusClientCircuitNotEstablished,
    StatusClientDangerousPort,
    StatusClientDangerousSocks,
    StatusClientSocksBadHostname,
    StatusGeneralBug,
    StatusGeneralClockJumped,
    StatusGeneralClockSkew,
    StatusGeneralDangerousVersion,
    StatusGeneralTooManyConnections,
    StatusServerAcceptedServerDescriptor,
    StatusServerBadServerDescriptor,
    StatusServerCheckingReachability,
    StatusServerExternalAddress,
    StatusServerHibernationStatus,
    StatusServerNameserverStatus,
    StatusServerReachabilityFailed,
    StatusServerReachabilitySucceeded,
    StreamClientProtocol,
    StreamCloseReason,
    StreamPurpose,
    StreamStatus,
    StreamTarget,
    TcpAddressPort,
)
from .types import (
    AnyAddress,
    AnyHost,
    AnyPort,
    Base16Bytes,
    Base32Bytes,
    Base64Bytes,
    BoolYesNo,
    DatetimeUTC,
    TimedeltaMilliseconds,
)
from .utils import (
    Message,
    MessageData,
    ReplySyntax,
    ReplySyntaxFlag,
    TrBeforeSetToNone,
    TrBeforeStringSplit,
    TrCast,
)

if TYPE_CHECKING:
    # The following line is needed so sphinx can get EventConfChanged right.
    from .reply import ReplyMapType  # noqa: F401

logger = logging.getLogger(__package__)


class EventWordInternal(StrEnum):
    """All events handled internally in this library."""

    #: The controller has been disconnected from Tor.
    #:
    #: See Also:
    #:     :class:`EventDisconnect`
    DISCONNECT = 'DISCONNECT'


class EventWord(StrEnum):
    """All possible events to subscribe to."""

    #: Circuit status changed.
    #:
    #: See Also:
    #:     :class:`EventCirc`
    CIRC = 'CIRC'

    #: Stream status changed.
    #:
    #: See Also:
    #:     :class:`EventStream`
    STREAM = 'STREAM'

    #: OR Connection status changed.
    #:
    #: See Also:
    #:     :class:`EventOrConn`
    ORCONN = 'ORCONN'

    #: Bandwidth used in the last second.
    #:
    #: See Also:
    #:     :class:`EventBandwidth`
    BW = 'BW'

    #: Debug log message.
    #:
    #: See Also:
    #:     :class:`EventLogDebug`
    DEBUG = 'DEBUG'

    #: Info log message.
    #:
    #: See Also:
    #:     :class:`EventLogInfo`
    INFO = 'INFO'

    #: Notice log message.
    #:
    #: See Also:
    #:     :class:`EventLogNotice`
    NOTICE = 'NOTICE'

    #: Warning log message.
    #:
    #: See Also:
    #:     :class:`EventLogWarn`
    WARN = 'WARN'

    #: Error log message.
    #:
    #: See Also:
    #:     :class:`EventLogErr`
    ERR = 'ERR'

    #: New descriptors available.
    #:
    #: See Also:
    #:     :class:`EventNewDesc`
    NEWDESC = 'NEWDESC'

    #: New Address mapping.
    #:
    #: See Also:
    #:     :class:`EventAddrMap`
    ADDRMAP = 'ADDRMAP'

    #: Descriptors uploaded to us in our role as authoritative dirserver.
    #:
    #: This event has been deprecated since Tor v0.3.3.1.
    AUTHDIR_NEWDESCS = 'AUTHDIR_NEWDESCS'

    #: Our descriptor changed.
    #:
    #: See Also:
    #:     :class:`EventDescChanged`
    DESCCHANGED = 'DESCCHANGED'

    #: General status event.
    #:
    #: See Also:
    #:     :class:`EventStatusGeneral`
    STATUS_GENERAL = 'STATUS_GENERAL'

    #: Client status event.
    #:
    #: See Also:
    #:     :class:`EventStatusClient`
    STATUS_CLIENT = 'STATUS_CLIENT'

    #: Server status event.
    #:
    #: See Also:
    #:     :class:`EventStatusServer`
    STATUS_SERVER = 'STATUS_SERVER'

    #: Our set of guard nodes has changed.
    #:
    #: See Also:
    #:     :class:`EventGuard`
    GUARD = 'GUARD'

    #: Network status has changed.
    #:
    #: See Also:
    #:     :class:`EventNetworkStatus`
    NS = 'NS'

    #: Bandwidth used on an application stream.
    #:
    #: See Also:
    #:     :class:`EventStreamBW`
    STREAM_BW = 'STREAM_BW'

    #: Per-country client stats.
    #:
    #: See Also:
    #:     :class:`EventClientsSeen`
    CLIENTS_SEEN = 'CLIENTS_SEEN'

    #: New consensus networkstatus has arrived.
    #:
    #: See Also:
    #:     :class:`EventNewConsensus`
    NEWCONSENSUS = 'NEWCONSENSUS'

    #: New circuit buildtime has been set.
    #:
    #: See Also:
    #:     :class:`EventBuildTimeoutSet`
    BUILDTIMEOUT_SET = 'BUILDTIMEOUT_SET'

    #: Signal received.
    #:
    #: See Also:
    #:     :class:`EventSignal`
    SIGNAL = 'SIGNAL'

    #: Configuration changed.
    #:
    #: See Also:
    #:     :class:`EventConfChanged`
    CONF_CHANGED = 'CONF_CHANGED'

    #: Circuit status changed slightly.
    #:
    #: See Also:
    #:     :class:`EventCircMinor`
    CIRC_MINOR = 'CIRC_MINOR'

    #: Pluggable transport launched.
    #:
    #: See Also:
    #:     :class:`EventTransportLaunched`
    TRANSPORT_LAUNCHED = 'TRANSPORT_LAUNCHED'

    #: Bandwidth used on an OR or DIR or EXIT connection.
    CONN_BW = 'CONN_BW'

    #: Bandwidth used by all streams attached to a circuit.
    #:
    #: See Also:
    #:     :class:`EventCircBW`
    CIRC_BW = 'CIRC_BW'

    #: Per-circuit cell stats.
    #:
    #: See Also:
    #:     :class:`EventCellStats`
    CELL_STATS = 'CELL_STATS'

    #: Token buckets refilled.
    #:
    #: See Also:
    #:     :class:`EventTbEmpty`
    TB_EMPTY = 'TB_EMPTY'

    #: HiddenService descriptors.
    #:
    #: See Also:
    #:     :class:`EventHsDesc`
    HS_DESC = 'HS_DESC'

    #: HiddenService descriptors content.
    #:
    #: See Also:
    #:     :class:`EventHsDescContent`
    HS_DESC_CONTENT = 'HS_DESC_CONTENT'

    #: Network liveness has changed.
    #:
    #: See Also:
    #:     :class:`EventNetworkLiveness`
    NETWORK_LIVENESS = 'NETWORK_LIVENESS'

    #: Pluggable Transport Logs.
    #:
    #: See Also:
    #:     :class:`EventPtLog`
    PT_LOG = 'PT_LOG'

    #: Pluggable Transport Status.
    #:
    #: See Also:
    #:     :class:`EventPtStatus`
    PT_STATUS = 'PT_STATUS'


@dataclass(kw_only=True, slots=True)
class Event(ABC):
    """Base class for all events."""

    #: Cached adapter used while deserializing the message.
    ADAPTER: ClassVar[TypeAdapter[Self] | None] = None

    #: Type of event this class is a parser for.
    TYPE: ClassVar[EventWordInternal | EventWord | None]

    @classmethod
    def adapter(cls) -> TypeAdapter[Self]:
        """Get a cached type adapter to deserialize a reply."""
        if cls.ADAPTER is None:
            cls.ADAPTER = TypeAdapter(cls)
        return cls.ADAPTER

    @classmethod
    @abstractmethod
    def from_message(cls, message: Message) -> Self:
        """Build an event structure from a received message."""


@dataclass(kw_only=True, slots=True)
class EventSimple(Event):
    """An event with a simple single syntax parser."""

    #: Simple syntax to parse the message from.
    SYNTAX: ClassVar[ReplySyntax]

    @classmethod
    def from_message(cls, message: Message) -> Self:
        """Build an event from a received message."""
        result = cls.SYNTAX.parse(message)
        return cls.adapter().validate_python(result)


@dataclass(kw_only=True, slots=True)
class EventDisconnect(Event):
    """
    Structure for a :attr:`~EventWordInternal.DISCONNECT` event.

    Note:
        This event is internal to ``aiostem``.

    """

    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWordInternal.DISCONNECT

    @classmethod
    def from_message(cls, message: Message) -> Self:
        """Build an event from a received message."""
        return cls.adapter().validate_python({})


@dataclass(kw_only=True, slots=True)
class EventAddrMap(EventSimple):
    """
    Structure for a :attr:`~EventWord.ADDRMAP` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#ADDRMAP

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=3,
        args_map=(None, 'original', 'replacement'),
        kwargs_map={
            None: 'expires_local',
            'error': 'error',
            'EXPIRES': 'expires',
            'CACHED': 'cached',
            'STREAMID': 'stream',
        },
        flags=ReplySyntaxFlag.KW_ENABLE
        | ReplySyntaxFlag.KW_QUOTED
        | ReplySyntaxFlag.KW_OMIT_KEYS,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.ADDRMAP

    #: Original address to replace.
    # Union is used around AnyHost to fix a weird bug with typing.get_type_hints().
    original: Union[AnyHost]  # noqa: UP007
    #: Replacement address, ``<error>`` is mapped to None.
    replacement: Annotated[Union[AnyHost, None], TrBeforeSetToNone({'<error>'})]  # noqa: UP007
    #: When this entry expires as an UTC date.
    expires: DatetimeUTC | None = None
    #: Error message when replacement is :obj:`None`.
    error: str | None = None
    #: Whether this value has been kept in cache.
    #:
    #: See Also:
    #:    https://docs.pydantic.dev/latest/api/standard_library_types/#booleans
    cached: BoolYesNo | None = None
    #: Stream identifier.
    stream: int | None = None


@dataclass(kw_only=True, slots=True)
class EventDescChanged(EventSimple):
    """
    Structure for a :attr:`~EventWord.DESCCHANGED` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#DESCCHANGED

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(args_min=1, args_map=(None,))
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.DESCCHANGED


@dataclass(kw_only=True, slots=True)
class EventStreamBW(EventSimple):
    """
    Structure for a :attr:`~EventWord.STREAM_BW` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#STREAM_BW

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=5,
        args_map=(None, 'stream', 'written', 'read', 'time'),
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.STREAM_BW

    #: Stream identifier.
    stream: int
    #: Number of bytes read by the application since the last event on this stream.
    read: int
    #: Number of bytes written by the application since the last event on this stream.
    written: int
    #: Records when Tor created the bandwidth event.
    time: DatetimeUTC


#: Describes a list of cell statistics for :class:`EventCellStats`.
GenericStatsMap: TypeAlias = Annotated[
    Mapping[str, int],
    BeforeValidator(dict),
    TrCast(
        Annotated[
            Sequence[
                Annotated[
                    tuple[str, str],
                    TrBeforeStringSplit(maxsplit=1, separator='='),
                ]
            ],
            TrBeforeStringSplit(separator=','),
        ],
        mode='before',
    ),
]


@dataclass(kw_only=True, slots=True)
class EventCirc(EventSimple):
    """
    Structure for a :attr:`~EventWord.CIRC` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#CIRC

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=3,
        args_map=(None, 'circuit', 'status'),
        kwargs_map={
            None: 'path',
            'BUILD_FLAGS': 'build_flags',
            'HS_POW': 'hs_pow',
            'HS_STATE': 'hs_state',
            'PURPOSE': 'purpose',
            'REASON': 'reason',
            'REMOTE_REASON': 'remote_reason',
            'REND_QUERY': 'rend_query',
            'SOCKS_USERNAME': 'socks_username',
            'SOCKS_PASSWORD': 'socks_password',
            'TIME_CREATED': 'time_created',
        },
        flags=ReplySyntaxFlag.KW_ENABLE
        | ReplySyntaxFlag.KW_OMIT_KEYS
        | ReplySyntaxFlag.KW_QUOTED,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.CIRC

    #: Circuit identifier this event is triggered for.
    circuit: NonNegativeInt
    #: Circuit status reported by this event.
    status: Annotated[CircuitStatus | str, Field(union_mode='left_to_right')]

    #: Circuit build flags.
    build_flags: (
        Annotated[
            AbstractSet[
                Annotated[
                    CircuitBuildFlags | str,
                    Field(union_mode='left_to_right'),
                ],
            ],
            TrBeforeStringSplit(),
        ]
        | None
    ) = None

    #: Hidden service proof of work effort attached to this circuit.
    hs_pow: (
        Annotated[
            CircuitHiddenServicePow,
            TrBeforeStringSplit(
                dict_keys=('type', 'effort'),
                separator=',',
            ),
        ]
        | None
    ) = None

    # Current hidden service state when applicable.
    hs_state: (
        Annotated[
            CircuitHiddenServiceState | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None

    #: List of servers in this circuit, when provided.
    path: Annotated[Sequence[LongServerName], TrBeforeStringSplit()] | None = None

    #: Current circuit purpose.
    purpose: (
        Annotated[
            CircuitPurpose | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None

    #: This field is provided only for ``FAILED`` and ``CLOSED`` events.
    reason: (
        Annotated[
            CircuitCloseReason | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None

    #: Provided only when we receive a DESTROY cell or RELAY_TRUNCATE message.
    remote_reason: (
        Annotated[
            CircuitCloseReason | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None

    #: Onion address related to this circuit (if any).
    rend_query: HiddenServiceAddress | None = None
    #: Username used by a SOCKS client to connect to Tor and initiate this circuit.
    socks_username: str | None = None
    #: Password used by a SOCKS client to connect to Tor and initiate this circuit.
    socks_password: str | None = None
    #: When this circuit was created, if provided.
    time_created: DatetimeUTC | None = None


@dataclass(kw_only=True, slots=True)
class EventStream(EventSimple):
    """
    Structure for a :attr:`~EventWord.STREAM` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#STREAM

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=5,
        args_map=(None, 'stream', 'status', 'circuit', 'target'),
        kwargs_map={
            'CLIENT_PROTOCOL': 'client_protocol',
            'ISO_FIELDS': 'iso_fields',
            'NYM_EPOCH': 'nym_epoch',
            'PURPOSE': 'purpose',
            'REASON': 'reason',
            'REMOTE_REASON': 'remote_reason',
            'SESSION_GROUP': 'session_group',
            'SOURCE': 'source',
            'SOURCE_ADDR': 'source_addr',
            'SOCKS_USERNAME': 'socks_username',
            'SOCKS_PASSWORD': 'socks_password',
        },
        flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.STREAM

    #: Stream identifier reported in this event.
    stream: int
    #: Status of the stream event reported here.
    status: StreamStatus
    #: Circuit ID linked to this stream.
    circuit: NonNegativeInt
    #: Target address or server of this stream (or ``0`` when unattached).
    target: StreamTarget

    #: The protocol that was used by a client to initiate this stream.
    client_protocol: (
        Annotated[
            StreamClientProtocol | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None

    #: Indicates the set of STREAM event fields for which stream isolation is enabled.
    iso_fields: Annotated[AbstractSet[str], TrBeforeStringSplit()] | None = None
    #: Nym epoch that was active when a client initiated this stream.
    nym_epoch: NonNegativeInt | None = None
    #: Only for :attr:`~.StreamStatus.NEW` and :attr:`~.StreamStatus.NEWRESOLVE` events.
    purpose: Annotated[StreamPurpose | str, Field(union_mode='left_to_right')] | None = None

    #: Provided only for ``FAILED``, ``CLOSED``, and ``DETACHED`` events,
    reason: (
        Annotated[
            StreamCloseReason | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None

    #: Provided only when we receive a RELAY_END message.
    remote_reason: (
        Annotated[
            StreamCloseReason | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None

    #: Indicates the session group of the listener port that a client used for this stream.
    session_group: int | None = None
    #: Username used by a SOCKS client to connect to Tor and initiate this circuit.
    socks_username: str | None = None
    #: Password used by a SOCKS client to connect to Tor and initiate this circuit.
    socks_password: str | None = None
    #: Generally either ``CACHE`` or ``EXIT``, used with :attr:`.StreamStatus.REMAP`.
    source: Annotated[RemapSource | str, Field(union_mode='left_to_right')] | None = None
    #: Source (local) address.
    source_addr: TcpAddressPort | None = None


@dataclass(kw_only=True, slots=True)
class EventOrConn(EventSimple):
    """
    Structure for a :attr:`~EventWord.ORCONN` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#ORCONN

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=3,
        args_map=(None, 'server', 'status'),
        kwargs_map={
            'REASON': 'reason',
            'NCIRCS': 'circuit_count',
            'ID': 'conn_id',
        },
        flags=ReplySyntaxFlag.KW_ENABLE,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.ORCONN

    #: Onion router server name reported in this event.
    server: LongServerName
    #: Status of the connection to the onion router.
    status: OrConnStatus
    #: When :attr:`status` is ``FAILED`` or ``CLOSED``, this is the reason why.
    reason: Annotated[OrConnCloseReason | str, Field(union_mode='left_to_right')] | None = None
    #: Counts both established and pending circuits.
    circuit_count: NonNegativeInt | None = None
    #: Connection identifier.
    conn_id: int | None = None


@dataclass(kw_only=True, slots=True)
class EventBandwidth(EventSimple):
    """
    Structure for a :attr:`~EventWord.BW` event.

    Note:
        Documentation seems to tell that there are keyword arguments on this event
        but the real implementation does not show any of them.
        These are being ignored here for now.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#BW

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=3,
        args_map=(None, 'read', 'written'),
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.BW

    #: Total amount of bytes read.
    read: int
    #: Total amount of bytes written
    written: int


@dataclass(kw_only=True, slots=True)
class EventGuard(EventSimple):
    """
    Structure for a :attr:`~EventWord.GUARD` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#GUARD

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=4,
        args_map=(None, 'type', 'name', 'status'),
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.GUARD

    #: Type of guard node, should be ``ENTRY``.
    type: str
    #: Full server name of the guard node.
    name: LongServerName
    #: Status of the guard in our event.
    status: Annotated[GuardEventStatus | str, Field(union_mode='left_to_right')]


@dataclass(kw_only=True, slots=True)
class EventNewDesc(EventSimple):
    """
    Structure for a :attr:`~EventWord.NEWDESC` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#NEWDESC

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=1,
        args_map=(None,),
        kwargs_map={None: 'servers'},
        kwargs_multi={'servers'},
        flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_OMIT_KEYS,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.NEWDESC

    #: List of new server identifiers received.
    servers: Sequence[LongServerName] = field(default_factory=list)


@dataclass(kw_only=True, slots=True)
class EventClientsSeen(EventSimple):
    """
    Structure for a :attr:`~EventWord.CLIENTS_SEEN` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#CLIENTS_SEEN

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=1,
        args_map=(None,),
        kwargs_map={
            'TimeStarted': 'time',
            'CountrySummary': 'countries',
            'IPVersions': 'ip_versions',
        },
        flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.CLIENTS_SEEN

    #: When the reported summary counts started.
    time: DatetimeUTC
    #: A map of countries seen by Tor.
    countries: GenericStatsMap
    #: A map of ip versions encountered.
    ip_versions: GenericStatsMap


@dataclass(kw_only=True, slots=True)
class EventBaseNetworkStatus(Event):
    """Base class for network status events."""

    SYNTAX_P: ClassVar[ReplySyntax] = ReplySyntax(args_min=2, args_map=('policy', 'ports'))
    SYNTAX_R: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=8,
        args_map=(
            'nickname',
            'identity',
            'digest',
            'exp_date',
            'exp_time',
            'ip',
            'or_port',
            'dir_port',
        ),
    )
    SYNTAX_W: ClassVar[ReplySyntax] = ReplySyntax(
        kwargs_map={
            'Bandwidth': 'bandwidth',
            'Measured': 'bw_measured',
            'Unmeasured': 'bw_unmeasured',
        },
        flags=ReplySyntaxFlag.KW_ENABLE,
    )

    #: Raw content of the new network status.
    routers: Sequence[RouterStatus]

    @classmethod
    def parse_router_statuses(cls, body: str) -> Sequence[dict[str, Any]]:
        """
        Parse router statuses to raw structures.

        Args:
            body: raw text containing router statuses.

        Returns:
            A list of dictonaries required to build a router status sequence.

        """
        current = {}  # type: dict[str, Any]
        routers = []  # type: list[dict[str, Any]]
        for line in body.splitlines():
            if not len(line):
                continue

            key, *args = line.split(' ', maxsplit=1)
            match key:
                case 'a':
                    addresses = current.setdefault('addresses', [])
                    addresses.append(args[0])

                case 'p':
                    current['port_policy'] = cls.SYNTAX_P.parse_string(args[0])

                case 'r':
                    if current:
                        routers.append(current)

                    current = {}
                    current.update(cls.SYNTAX_R.parse_string(args[0]))

                case 's':
                    current['flags'] = args[0].split(' ')

                case 'w':
                    current.update(cls.SYNTAX_W.parse_string(args[0]))

                case _:  # pragma: no cover
                    content = args[0] if len(args) else '__NONE__'
                    logger.warning('Unhandled network status type %s: %s', key, content)

        if current:  # pragma: no branch
            routers.append(current)
        return routers

    @classmethod
    def from_message(cls, message: Message) -> Self:
        """Build an event from a received message."""
        if not len(message.items) or not isinstance(message.items[0], MessageData):
            msg = "Event 'NS' has no data attached to it!"
            raise ReplySyntaxError(msg)

        routers = cls.parse_router_statuses(message.items[0].data)
        return cls.adapter().validate_python({'routers': routers})


@dataclass(kw_only=True, slots=True)
class EventNetworkStatus(EventBaseNetworkStatus):
    """
    Structure for a :attr:`~EventWord.NS` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#NS

    """

    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.NS


@dataclass(kw_only=True, slots=True)
class EventNewConsensus(EventBaseNetworkStatus):
    """
    Structure for a :attr:`~EventWord.NEWCONSENSUS` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#NEWCONSENSUS

    """

    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.NEWCONSENSUS


@dataclass(kw_only=True, slots=True)
class EventBuildTimeoutSet(EventSimple):
    """
    Structure for a :attr:`~EventWord.BUILDTIMEOUT_SET` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#BUILDTIMEOUT_SET

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=2,
        args_map=(None, 'type'),
        kwargs_map={
            'TOTAL_TIMES': 'total_times',
            'TIMEOUT_MS': 'timeout_ms',
            'XM': 'xm',
            'ALPHA': 'alpha',
            'CUTOFF_QUANTILE': 'cutoff_quantile',
            'TIMEOUT_RATE': 'timeout_rate',
            'CLOSE_MS': 'close_ms',
            'CLOSE_RATE': 'close_rate',
        },
        flags=ReplySyntaxFlag.KW_ENABLE,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.BUILDTIMEOUT_SET

    #: Type of event we just received.
    type: Literal['COMPUTED', 'RESET', 'SUSPENDED', 'DISCARD', 'RESUME']
    #: Integer count of timeouts stored.
    total_times: NonNegativeInt
    #: Integer timeout in milliseconds.
    timeout_ms: TimedeltaMilliseconds
    #: Estimated integer Pareto parameter Xm in milliseconds.
    xm: TimedeltaMilliseconds
    #: Estimated floating point Paredo parameter alpha.
    alpha: float
    #: Floating point CDF quantile cutoff point for this timeout.
    cutoff_quantile: float
    #: Floating point ratio of circuits that timeout.
    timeout_rate: float
    #: How long to keep measurement circs in milliseconds.
    close_ms: TimedeltaMilliseconds
    #: Floating point ratio of measurement circuits that are closed.
    close_rate: float


@dataclass(kw_only=True, slots=True)
class EventSignal(EventSimple):
    """
    Structure for a :attr:`~EventWord.SIGNAL` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#SIGNAL

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(args_min=2, args_map=(None, 'signal'))
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.SIGNAL

    #: The signal received by Tor.
    signal: Signal


@dataclass(kw_only=True, slots=True)
class EventCircBW(EventSimple):
    """
    Structure for a :attr:`~EventWord.CIRC_BW` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#CIRC_BW

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=1,
        args_map=(None,),
        kwargs_map={
            'ID': 'circuit',
            'TIME': 'time',
            'READ': 'read',
            'DELIVERED_READ': 'read_delivered',
            'OVERHEAD_READ': 'read_overhead',
            'WRITTEN': 'written',
            'DELIVERED_WRITTEN': 'written_delivered',
            'OVERHEAD_WRITTEN': 'written_overhead',
            'SS': 'slow_start',
            'CWND': 'cwnd',
            'RTT': 'rtt',
            'MIN_RTT': 'rtt_min',
        },
        flags=ReplySyntaxFlag.KW_ENABLE,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.CIRC_BW

    #: Records when Tor created the bandwidth event.
    time: DatetimeUTC

    #: Number of bytes read on this circuit since the last :attr:`~EventWord.CIRC_BW` event.
    read: int
    #: Byte count for incoming delivered relay messages.
    read_delivered: int
    #: Overhead of extra unused bytes at the end of read messages.
    read_overhead: int

    #: Number of bytes written on this circuit since the last :attr:`~EventWord.CIRC_BW` event.
    written: int
    #: Byte count for outgoing delivered relay messages.
    written_delivered: int
    #: Overhead of extra unused bytes at the end of written messages.
    written_overhead: int

    #: Provides an indication if the circuit is in slow start or not.
    slow_start: bool | None = None
    #: Size of the congestion window in terms of number of cells.
    cwnd: int | None = None
    #: The ``N_EWMA`` smoothed current RTT value.
    rtt: TimedeltaMilliseconds | None = None
    #: Minimum RTT value of the circuit.
    rtt_min: TimedeltaMilliseconds | None = None


@dataclass(kw_only=True, slots=True)
class EventConfChanged(Event, ReplyGetMap):
    """
    Structure for a :attr:`~EventWord.CONF_CHANGED` event.

    Hint:
        This class behaves somehow like :class:`.ReplyGetConf`.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#CONF_CHANGED

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        flags=(
            ReplySyntaxFlag.KW_ENABLE
            | ReplySyntaxFlag.KW_OMIT_VALS
            | ReplySyntaxFlag.KW_EXTRA
            | ReplySyntaxFlag.KW_RAW
        )
    )

    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.CONF_CHANGED

    @classmethod
    def from_message(cls, message: Message) -> Self:
        """Build an event dataclass from a received message."""
        result = {}  # type: dict[str, Any]
        if len(message.items) > 1:
            result['data'] = cls._key_value_extract(message.items[1:])
        return cls.adapter().validate_python(result)


@dataclass(kw_only=True, slots=True)
class EventCircMinor(EventSimple):
    """
    Structure for a :attr:`~EventWord.CIRC_MINOR` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#CIRC_MINOR

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=3,
        args_map=(None, 'circuit', 'event'),
        kwargs_map={
            None: 'path',
            'BUILD_FLAGS': 'build_flags',
            'HS_STATE': 'hs_state',
            'PURPOSE': 'purpose',
            'REND_QUERY': 'rend_query',
            'TIME_CREATED': 'time_created',
            'OLD_HS_STATE': 'old_hs_state',
            'OLD_PURPOSE': 'old_purpose',
        },
        flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_OMIT_KEYS,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.CIRC_MINOR

    #: Circuit identifier.
    circuit: NonNegativeInt
    #: Circuit event, either ``PURPOSE_CHANGED`` or ``CANNIBALIZED``.
    event: Annotated[CircuitEvent | str, Field(union_mode='left_to_right')]
    #: Circuit path, when provided.
    path: Annotated[Sequence[LongServerName], TrBeforeStringSplit()] | None = None

    #: Circuit build flags.
    build_flags: (
        Annotated[
            AbstractSet[
                Annotated[
                    CircuitBuildFlags | str,
                    Field(union_mode='left_to_right'),
                ],
            ],
            TrBeforeStringSplit(),
        ]
        | None
    ) = None

    # Current hidden service state when applicable.
    hs_state: (
        Annotated[
            CircuitHiddenServiceState | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None
    #: When this circuit was created.
    time_created: DatetimeUTC | None = None

    #: Current circuit purpose.
    purpose: (
        Annotated[
            CircuitPurpose | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None

    #: Onion address related to this circuit.
    rend_query: HiddenServiceAddress | None = None

    #: Previous hidden service state when applicable.
    old_hs_state: (
        Annotated[
            CircuitHiddenServiceState | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None
    #: Previous circuit purpose.
    old_purpose: (
        Annotated[
            CircuitPurpose | str,
            Field(union_mode='left_to_right'),
        ]
        | None
    ) = None


#: Describes a list of cell statistics for :class:`EventCellStats`.
CellsByType: TypeAlias = Annotated[
    Mapping[str, int],
    BeforeValidator(dict),
    TrCast(
        Annotated[
            Sequence[
                Annotated[
                    tuple[str, str],
                    TrBeforeStringSplit(maxsplit=1, separator=':'),
                ]
            ],
            TrBeforeStringSplit(separator=','),
        ],
        mode='before',
    ),
]

#: Describes a list of cell time statistics for :class:`EventCellStats`.
MsecByType: TypeAlias = Annotated[
    Mapping[str, TimedeltaMilliseconds],
    BeforeValidator(dict),
    TrCast(
        Annotated[
            Sequence[
                Annotated[
                    tuple[str, str],
                    TrBeforeStringSplit(maxsplit=1, separator=':'),
                ]
            ],
            TrBeforeStringSplit(separator=','),
        ],
        mode='before',
    ),
]


@dataclass(kw_only=True, slots=True)
class EventCellStats(EventSimple):
    """
    Structure for a :attr:`~EventWord.CELL_STATS` event.

    Important:
        These events are only generated if TestingTorNetwork is set.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#CELL_STATS

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=1,
        args_map=(None,),
        kwargs_map={
            'ID': 'circuit',
            'InboundConn': 'inbound_conn_id',
            'InboundQueue': 'inbound_queue',
            'InboundAdded': 'inbound_added',
            'InboundRemoved': 'inbound_removed',
            'InboundTime': 'inbound_time',
            'OutboundConn': 'outbound_conn_id',
            'OutboundQueue': 'outbound_queue',
            'OutboundAdded': 'outbound_added',
            'OutboundRemoved': 'outbound_removed',
            'OutboundTime': 'outbound_time',
        },
        flags=ReplySyntaxFlag.KW_ENABLE,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.CELL_STATS

    #: Circuit identifier only included if the circuit originates at this node.
    circuit: NonNegativeInt | None = None

    #: InboundQueue is the identifier of the inbound circuit queue of this circuit.
    inbound_queue: int | None = None
    #: Locally unique IDs of inbound OR connection.
    inbound_conn_id: int | None = None
    #: Total number of cells by cell type added to inbound queue.
    inbound_added: CellsByType | None = None
    #: Total number of cells by cell type processed from inbound queue.
    inbound_removed: CellsByType | None = None
    #: Total waiting times in milliseconds of all processed cells by cell type.
    inbound_time: MsecByType | None = None

    #: OutboundQueue is the identifier of the outbound circuit queue of this circuit.
    outbound_queue: int | None = None
    #: Locally unique IDs of outbound OR connection.
    outbound_conn_id: int | None = None
    #: Total number of cells by cell type added to outbound queue.
    outbound_added: CellsByType | None = None
    #: Total number of cells by cell type processed from outbound queue.
    outbound_removed: CellsByType | None = None
    #: Total waiting times in milliseconds of all processed cells by cell type.
    outbound_time: MsecByType | None = None


@dataclass(kw_only=True, slots=True)
class EventConnBW(EventSimple):
    """
    Structure for a :attr:`~EventWord.CONN_BW` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#CONN_BW

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=1,
        args_map=(None,),
        kwargs_map={
            'ID': 'conn_id',
            'TYPE': 'conn_type',
            'READ': 'read',
            'WRITTEN': 'written',
        },
        flags=ReplySyntaxFlag.KW_ENABLE,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.CONN_BW

    #: Identifier for this connection.
    conn_id: int
    #: Connection type, typically ``OR`` / ``DIR`` / ``EXIT``.
    conn_type: str
    #: Number of bytes read by Tor since the last event on this connection.
    read: int
    #: Number of bytes written by Tor since the last event on this connection.
    written: int


@dataclass(kw_only=True, slots=True)
class EventTbEmpty(EventSimple):
    """
    Structure for a :attr:`~EventWord.TB_EMPTY` event.

    Important:
        These events are only generated if TestingTorNetwork is set.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#TB_EMPTY

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=2,
        args_map=(None, 'bucket'),
        kwargs_map={
            'ID': 'conn_id',
            'LAST': 'last',
            'READ': 'read',
            'WRITTEN': 'written',
        },
        flags=ReplySyntaxFlag.KW_ENABLE,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.TB_EMPTY

    #: Name of the refilled bucket that was previously empty.
    bucket: Literal['GLOBAL', 'RELAY', 'ORCONN']

    #: Connection ID, only included when :attr:`bucket` is ``ORCONN``.
    conn_id: int | None = None

    #: Duration since the last refill.
    last: TimedeltaMilliseconds

    #: Duration that the read bucket was empty since the last refill.
    read: TimedeltaMilliseconds

    #: Duration that the write bucket was empty since the last refill.
    written: TimedeltaMilliseconds


@dataclass(kw_only=True, slots=True)
class EventHsDesc(EventSimple):
    """
    Structure for a :attr:`~EventWord.HS_DESC` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#HS_DESC

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=5,
        args_map=(None, 'action', 'address', 'auth_type', 'hs_dir'),
        kwargs_map={
            None: 'descriptor_id',
            'REASON': 'reason',
            'REPLICA': 'replica',
            'HSDIR_INDEX': 'hs_dir_index',
        },
        flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_OMIT_KEYS,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.HS_DESC

    #: Kind of action reported in this status update.
    action: HsDescAction
    #: Onion address the report status is for (without the ``.onion`` suffix).
    address: HiddenServiceAddress | Literal['UNKNOWN']
    #: Client authentication here is always :attr:`~.HsDescAuthTypeStr.NO_AUTH`.
    auth_type: HsDescAuthTypeStr
    #: The descriptor blinded key used for the index value at the ``HsDir``.
    descriptor_id: Base32Bytes | Base64Bytes | None = None
    #: Hidden service directory answering this request.
    hs_dir: LongServerName | Literal['UNKNOWN']
    #: Contains the computed index of the HsDir the descriptor was uploaded to or fetched from.
    hs_dir_index: Base16Bytes | None = None
    #: If :attr:`action` is :attr:`~.HsDescAction.FAILED`, Tor SHOULD send a reason field.
    reason: HsDescFailReason | None = None
    #: Field is not used for the :attr:`~.HsDescAction.CREATED` event because v3 doesn't use
    #: the replica number in the descriptor ID computation.
    replica: int | None = None


@dataclass(kw_only=True)
class EventHsDescContent(Event):
    """
    Structure for a :attr:`~EventWord.HS_DESC_CONTENT` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#HS_DESC_CONTENT

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=4,
        args_map=(None, 'address', 'descriptor_id', 'hs_dir'),
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.HS_DESC_CONTENT

    #: Onion address the report status is for (without the ``.onion`` suffix).
    address: HiddenServiceAddress | Literal['UNKNOWN']
    #: Hidden service directory answering this request.
    hs_dir: LongServerName | Literal['UNKNOWN']
    #: Unique identifier for the descriptor.
    descriptor_id: Base32Bytes | Base64Bytes | None = None
    #: Text content of the hidden service descriptor.
    descriptor_text: str

    @cached_property
    def descriptor(self) -> HsDescV2 | HsDescV3:
        """Get the parsed descriptor."""
        match self.address:
            case HiddenServiceAddressV2():
                return HsDescV2.from_text(self.descriptor_text)

            case HiddenServiceAddressV3():
                return HsDescV3.from_text(self.descriptor_text)

            case _:  # pragma: no cover
                msg = 'Unhandled hidden service descriptor format.'
                raise NotImplementedError(msg)

    @classmethod
    def from_message(cls, message: Message) -> Self:
        """Build an event from a received message."""
        if not len(message.items) or not isinstance(message.items[0], MessageData):
            msg = "Event 'HS_DESC_CONTENT' has no data attached to it!"
            raise ReplySyntaxError(msg)

        result = cls.SYNTAX.parse(message.items[0])
        descriptor = message.items[0].data
        return cls.adapter().validate_python({**result, 'descriptor_text': descriptor})


@dataclass(kw_only=True, slots=True)
class EventNetworkLiveness(EventSimple):
    """
    Structure for a :attr:`~EventWord.NETWORK_LIVENESS` event.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#NETWORK_LIVENESS

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=2,
        args_map=(None, 'status'),
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.NETWORK_LIVENESS

    #: Current network status.
    status: LivenessStatus


@dataclass(kw_only=True, slots=True)
class EventLog(Event):
    """
    Base class for any event log.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#LOG

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=1,
        args_map=('severity', 'message'),
        flags=ReplySyntaxFlag.POS_REMAIN,
    )

    #: Log severity.
    severity: LogSeverity
    #: Log message.
    message: str

    @classmethod
    def from_message(cls, message: Message) -> Self:
        """Build an event from a received message."""
        result = {}  # type: dict[str, Any]
        if len(message.items) and isinstance(message.items[0], MessageData):
            result.update(cls.SYNTAX.parse(message.items[0]))
            result['message'] = message.items[0].data
        else:
            result.update(cls.SYNTAX.parse(message))

        return cls.adapter().validate_python(result)


@dataclass(kw_only=True, slots=True)
class EventLogDebug(EventLog):
    """
    Event parser for :attr:`~EventWord.DEBUG` events.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#LOG

    """

    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.DEBUG


@dataclass(kw_only=True, slots=True)
class EventLogInfo(EventLog):
    """
    Event parser for :attr:`~EventWord.INFO` events.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#LOG

    """

    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.INFO


@dataclass(kw_only=True, slots=True)
class EventLogNotice(EventLog):
    """
    Event parser for :attr:`~EventWord.NOTICE` events.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#LOG

    """

    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.NOTICE


@dataclass(kw_only=True, slots=True)
class EventLogWarn(EventLog):
    """
    Event parser for :attr:`~EventWord.WARN` events.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#LOG

    """

    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.WARN


@dataclass(kw_only=True, slots=True)
class EventLogErr(EventLog):
    """
    Event parser for :attr:`~EventWord.ERR` events.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#LOG

    """

    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.ERR


def _discriminate_status_by_action(v: Any) -> str:
    """
    Discriminate a `STATUS_*` event by its actions.

    Args:
        v: The raw value to serialize/deserialize the event.

    Returns:
        The tag corresponding to the structure to parse in the ``arguments`` union.

    """
    match v:
        case Mapping():
            return v['action']
        case None:
            return '__NONE__'
        case _:  # pragma: no cover
            return v.action


@dataclass(kw_only=True, slots=True)
class EventStatus(Event):
    """
    Base class for all ``STATUS_*`` events.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#STATUS

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=3,
        args_map=(None, 'severity', 'action', 'argstring'),
        flags=ReplySyntaxFlag.POS_REMAIN,
    )
    SUBSYNTAXES: ClassVar[Mapping[str, ReplySyntax | None]]

    #: Severity of the reported status.
    severity: Annotated[
        Literal[LogSeverity.NOTICE, LogSeverity.WARNING, LogSeverity.ERROR],
        LogSeverity,
    ]
    #: Status action reported by this event (sub-classed).
    action: StrEnum

    @classmethod
    def from_message(cls, message: Message) -> Self:
        """Build an event from a received message."""
        result = {'arguments': None}  # type: dict[str, Any]
        result.update(cls.SYNTAX.parse(message))

        argstring = result.pop('argstring', '')
        action = result['action']
        if action in cls.SUBSYNTAXES:
            syntax = cls.SUBSYNTAXES[action]
            if syntax is not None:
                # `action` here is used as a discriminator.
                arguments = {'action': action}  # type: dict[str, Any]
                arguments.update(syntax.parse_string(argstring))
                result['arguments'] = arguments
        else:
            logger.info("No syntax handler for action '%s'.", action)
        return cls.adapter().validate_python(result)


@dataclass(kw_only=True, slots=True)
class EventStatusGeneral(EventStatus):
    """
    Event parser for :attr:`~EventWord.STATUS_GENERAL` events.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#STATUS

    """

    SUBSYNTAXES: ClassVar[Mapping[str, ReplySyntax | None]] = {
        'BUG': ReplySyntax(
            kwargs_map={'REASON': 'reason'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'CLOCK_JUMPED': ReplySyntax(
            kwargs_map={'TIME': 'time'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'CLOCK_SKEW': ReplySyntax(
            kwargs_map={
                'SOURCE': 'source',
                'SKEW': 'skew',
            },
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'DANGEROUS_VERSION': ReplySyntax(
            kwargs_map={
                'CURRENT': 'current',
                'REASON': 'reason',
                'RECOMMENDED': 'recommended',
            },
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'DIR_ALL_UNREACHABLE': None,
        'TOO_MANY_CONNECTIONS': ReplySyntax(
            kwargs_map={'CURRENT': 'current'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
    }
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.STATUS_GENERAL

    #: Which action this general status event is for.
    action: StatusActionGeneral

    #: Arguments associated with the :attr:`action`.
    arguments: Annotated[
        Union[  # noqa: UP007
            Annotated[StatusGeneralBug, Tag('BUG')],
            Annotated[StatusGeneralClockJumped, Tag('CLOCK_JUMPED')],
            Annotated[StatusGeneralClockSkew, Tag('CLOCK_SKEW')],
            Annotated[StatusGeneralDangerousVersion, Tag('DANGEROUS_VERSION')],
            Annotated[StatusGeneralTooManyConnections, Tag('TOO_MANY_CONNECTIONS')],
            Annotated[None, Tag('__NONE__')],
        ],
        Discriminator(_discriminate_status_by_action),
    ]


@dataclass(kw_only=True, slots=True)
class EventStatusClient(EventStatus):
    """
    Event parser for :attr:`~EventWord.STATUS_CLIENT` events.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#STATUS

    """

    SUBSYNTAXES: ClassVar[Mapping[str, ReplySyntax | None]] = {
        'BOOTSTRAP': ReplySyntax(
            kwargs_map={
                'COUNT': 'count',
                'HOST': 'host',
                'HOSTADDR': 'hostaddr',
                'PROGRESS': 'progress',
                'REASON': 'reason',
                'RECOMMENDATION': 'recommendation',
                'SUMMARY': 'summary',
                'TAG': 'tag',
                'WARNING': 'warning',
            },
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'ENOUGH_DIR_INFO': None,
        'NOT_ENOUGH_DIR_INFO': None,
        'CIRCUIT_ESTABLISHED': None,
        'CIRCUIT_NOT_ESTABLISHED': ReplySyntax(
            kwargs_map={'REASON': 'reason'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'CONSENSUS_ARRIVED': None,
        'DANGEROUS_PORT': ReplySyntax(
            kwargs_map={
                'PORT': 'port',
                'RESULT': 'result',
            },
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'DANGEROUS_SOCKS': ReplySyntax(
            kwargs_map={
                'PROTOCOL': 'protocol',
                'ADDRESS': 'address',
            },
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'SOCKS_UNKNOWN_PROTOCOL': None,
        'SOCKS_BAD_HOSTNAME': ReplySyntax(
            kwargs_map={'HOSTNAME': 'hostname'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
    }
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.STATUS_CLIENT

    #: Which action this client status event is for.
    action: StatusActionClient

    #: Arguments associated with the :attr:`action`.
    arguments: Annotated[
        Union[  # noqa: UP007
            Annotated[StatusClientBootstrap, Tag('BOOTSTRAP')],
            Annotated[StatusClientCircuitNotEstablished, Tag('CIRCUIT_NOT_ESTABLISHED')],
            Annotated[StatusClientDangerousPort, Tag('DANGEROUS_PORT')],
            Annotated[StatusClientDangerousSocks, Tag('DANGEROUS_SOCKS')],
            Annotated[StatusClientSocksBadHostname, Tag('SOCKS_BAD_HOSTNAME')],
            Annotated[None, Tag('__NONE__')],
        ],
        Discriminator(_discriminate_status_by_action),
    ]


@dataclass(kw_only=True, slots=True)
class EventStatusServer(EventStatus):
    """
    Event parser for :attr:`~EventWord.STATUS_SERVER` events.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#STATUS

    """

    SUBSYNTAXES: ClassVar[Mapping[str, ReplySyntax | None]] = {
        'EXTERNAL_ADDRESS': ReplySyntax(
            kwargs_map={
                'ADDRESS': 'address',
                'HOSTNAME': 'hostname',
                'METHOD': 'method',
            },
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'CHECKING_REACHABILITY': ReplySyntax(
            kwargs_map={'ORADDRESS': 'or_address'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'REACHABILITY_SUCCEEDED': ReplySyntax(
            kwargs_map={'ORADDRESS': 'or_address'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'GOOD_SERVER_DESCRIPTOR': None,
        'NAMESERVER_STATUS': ReplySyntax(
            kwargs_map={
                'NS': 'ns',
                'STATUS': 'status',
                'ERR': 'err',
            },
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'NAMESERVER_ALL_DOWN': None,
        'DNS_HIJACKED': None,
        'DNS_USELESS': None,
        'BAD_SERVER_DESCRIPTOR': ReplySyntax(
            kwargs_map={
                'DIRAUTH': 'dir_auth',
                'REASON': 'reason',
            },
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'ACCEPTED_SERVER_DESCRIPTOR': ReplySyntax(
            kwargs_map={'DIRAUTH': 'dir_auth'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'REACHABILITY_FAILED': ReplySyntax(
            kwargs_map={'ORADDRESS': 'or_address'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
        'HIBERNATION_STATUS': ReplySyntax(
            kwargs_map={'STATUS': 'status'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        ),
    }
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.STATUS_SERVER

    #: Which action this server status event is for.
    action: StatusActionServer

    #: Arguments associated with the :attr:`action`.
    arguments: Annotated[
        Union[  # noqa: UP007
            Annotated[StatusServerExternalAddress, Tag('EXTERNAL_ADDRESS')],
            Annotated[StatusServerCheckingReachability, Tag('CHECKING_REACHABILITY')],
            Annotated[StatusServerReachabilitySucceeded, Tag('REACHABILITY_SUCCEEDED')],
            Annotated[StatusServerNameserverStatus, Tag('NAMESERVER_STATUS')],
            Annotated[StatusServerBadServerDescriptor, Tag('BAD_SERVER_DESCRIPTOR')],
            Annotated[StatusServerAcceptedServerDescriptor, Tag('ACCEPTED_SERVER_DESCRIPTOR')],
            Annotated[StatusServerReachabilityFailed, Tag('REACHABILITY_FAILED')],
            Annotated[StatusServerHibernationStatus, Tag('HIBERNATION_STATUS')],
            Annotated[None, Tag('__NONE__')],
        ],
        Discriminator(_discriminate_status_by_action),
    ]


@dataclass(kw_only=True, slots=True)
class EventTransportLaunched(EventSimple):
    """
    Event parser for :attr:`~EventWord.TRANSPORT_LAUNCHED` events.

    See Also:
        https://spec.torproject.org/control-spec/replies.html#TRANSPORT_LAUNCHED

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=5,
        args_map=(None, 'side', 'name', 'host', 'port'),
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.TRANSPORT_LAUNCHED

    #: Which side the transport was launched for.
    side: Literal['client', 'server']
    #: Name of the pluggable transport.
    name: str
    #: Host hosting the pluggable transport.
    host: AnyAddress
    #: Associated TCP port.
    port: AnyPort


@dataclass(kw_only=True, slots=True)
class EventPtLog(EventSimple):
    """
    Event parser for :attr:`~EventWord.PT_LOG` events.

    See Also:
        - https://spec.torproject.org/control-spec/replies.html#PT_LOG
        - https://spec.torproject.org/pt-spec/ipc.html#log-messages

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=1,
        args_map=(None,),
        kwargs_map={
            'PT': 'program',
            'MESSAGE': 'message',
            'SEVERITY': 'severity',
        },
        flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED | ReplySyntaxFlag.KW_EXTRA,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.PT_LOG

    #: Program path as defined in the ``TransportPlugin`` configuration option.
    program: str
    #: The status message that the PT sends back to the tor parent minus
    #: the ``STATUS`` string prefix.
    message: str
    #: Log severity.
    severity: LogSeverity


@dataclass(kw_only=True, slots=True)
class EventPtStatus(Event):
    """
    Event parser for :attr:`~EventWord.PT_STATUS` events.

    See Also:
        - https://spec.torproject.org/control-spec/replies.html#PT_STATUS
        - https://spec.torproject.org/pt-spec/ipc.html#status-messages

    """

    SYNTAX: ClassVar[ReplySyntax] = ReplySyntax(
        args_min=1,
        args_map=(None,),
        kwargs_map={
            'TRANSPORT': 'transport',
            'PT': 'program',
        },
        flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED | ReplySyntaxFlag.KW_EXTRA,
    )
    TYPE: ClassVar[EventWordInternal | EventWord | None] = EventWord.PT_STATUS

    #: Program path as defined in the ``TransportPlugin`` configuration option.
    program: str
    #: This value indicates a hint on what the PT is such as the name or the protocol used.
    transport: str
    #: All keywords reported by the underlying PT plugin, such as messages, etc...
    values: Mapping[str, str] = field(default_factory=dict)

    @classmethod
    def from_message(cls, message: Message) -> Self:
        """Build an event from a received message."""
        extract = dict(cls.SYNTAX.parse(message))
        result: dict[str, Any] = {
            key: extract.pop(key, None) for key in ('program', 'transport')
        }
        result['values'] = extract
        return cls.adapter().validate_python(result)


@dataclass(kw_only=True, slots=True)
class EventUnknown(Event):
    """
    Structure for an unknown event.

    This structure is the default fallback when no event class suits the event type
    the user subscribed to.

    """

    TYPE: ClassVar[EventWordInternal | EventWord | None] = None

    #: Original message received for this event.
    message: Message

    @classmethod
    def from_message(cls, message: Message) -> Self:
        """Build an event from a received message."""
        return cls.adapter().validate_python({'message': message})


_EVENT_MAP = {
    'ADDRMAP': EventAddrMap,
    'BUILDTIMEOUT_SET': EventBuildTimeoutSet,
    'BW': EventBandwidth,
    'DISCONNECT': EventDisconnect,
    'CIRC': EventCirc,
    'CONF_CHANGED': EventConfChanged,
    'CLIENTS_SEEN': EventClientsSeen,
    'CELL_STATS': EventCellStats,
    'CIRC_MINOR': EventCircMinor,
    'CIRC_BW': EventCircBW,
    'CONN_BW': EventConnBW,
    'DESCCHANGED': EventDescChanged,
    'GUARD': EventGuard,
    'HS_DESC': EventHsDesc,
    'HS_DESC_CONTENT': EventHsDescContent,
    'NETWORK_LIVENESS': EventNetworkLiveness,
    'NEWCONSENSUS': EventNewConsensus,
    'NEWDESC': EventNewDesc,
    'NS': EventNetworkStatus,
    'DEBUG': EventLogDebug,
    'INFO': EventLogInfo,
    'NOTICE': EventLogNotice,
    'WARN': EventLogWarn,
    'ERR': EventLogErr,
    'ORCONN': EventOrConn,
    'PT_LOG': EventPtLog,
    'PT_STATUS': EventPtStatus,
    'SIGNAL': EventSignal,
    'STATUS_GENERAL': EventStatusGeneral,
    'STATUS_CLIENT': EventStatusClient,
    'STATUS_SERVER': EventStatusServer,
    'STREAM': EventStream,
    'STREAM_BW': EventStreamBW,
    'TB_EMPTY': EventTbEmpty,
    'TRANSPORT_LAUNCHED': EventTransportLaunched,
}  # type: Mapping[str, type[Event]]


def event_from_message(message: Message) -> Event:
    """
    Parse an event message to the corresponding event structure.

    Args:
        message: An event message to parse.

    Raises:
        MessageError: When the message is not an event.

    Returns:
        A parsed event corresponding to the event in the provided message.

    """
    if not message.is_event:
        msg = 'The provided message is not an event!'
        raise MessageError(msg)

    handler = _EVENT_MAP.get(message.keyword)
    if handler is None:
        handler = EventUnknown
    return handler.from_message(message)



from __future__ import annotations

import secrets
from collections.abc import Mapping, MutableMapping, MutableSequence
from dataclasses import dataclass, field
from enum import StrEnum
from typing import TYPE_CHECKING, Annotated, Any, ClassVar, Literal, Self, Union

from pydantic import Discriminator, NonNegativeInt, Tag, TypeAdapter
from pydantic_core import core_schema

from .event import EventWord
from .exceptions import CommandError
from .structures import (
    CircuitPurpose,
    DescriptorPurpose,
    Feature,
    HiddenServiceAddress,
    HiddenServiceAddressV3,
    HsDescClientAuthV2,
    HsDescClientAuthV3,
    LongServerName,
    OnionClientAuthFlags,
    OnionClientAuthKey,
    OnionServiceFlags,
    OnionServiceKey,
    OnionServiceNewKey,
    OnionServiceNewKeyStruct,
    Signal,
    StreamCloseReasonInt,
    VirtualPort,
)
from .types import AnyHost, AnyPort, Base16Bytes, BoolYesNo
from .utils.argument import ArgumentKeyword, ArgumentString, QuoteStyle
from .utils.transformers import TrBeforeStringSplit

if TYPE_CHECKING:
    from pydantic import GetCoreSchemaHandler
    from pydantic_core.core_schema import CoreSchema, SerializerFunctionWrapHandler


class CommandWord(StrEnum):
    """All handled command words."""

    #: Change the value of one or more configuration variables.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.set_conf`
    #:     - Command implementation: :class:`CommandSetConf`
    #:     - Reply implementation: :class:`.ReplySetConf`
    SETCONF = 'SETCONF'

    #: Remove all settings for a given configuration option entirely.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.reset_conf`
    #:     - Command implementation: :class:`CommandResetConf`
    #:     - Reply implementation: :class:`.ReplyResetConf`
    RESETCONF = 'RESETCONF'

    #: Request the value of zero or more configuration variable(s).
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.get_conf`
    #:     - Command implementation: :class:`CommandGetConf`
    #:     - Reply implementation: :class:`.ReplyGetConf`
    GETCONF = 'GETCONF'

    #: Request the server to inform the client about interesting events.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.set_events`
    #:     - Command implementation: :class:`CommandSetEvents`
    #:     - Reply implementation: :class:`.ReplySetEvents`
    SETEVENTS = 'SETEVENTS'

    #: Used to authenticate to the server.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.authenticate`
    #:     - Command implementation: :class:`CommandAuthenticate`
    #:     - Reply implementation: :class:`.ReplyAuthenticate`
    AUTHENTICATE = 'AUTHENTICATE'

    #: Instructs the server to write out its config options into its ``torrc``.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.save_conf`
    #:     - Command implementation: :class:`CommandSaveConf`
    #:     - Reply implementation: :class:`.ReplySaveConf`
    SAVECONF = 'SAVECONF'

    #: Send a signal to the server.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.signal`
    #:     - Command implementation: :class:`CommandSignal`
    #:     - Reply implementation: :class:`.ReplySignal`
    SIGNAL = 'SIGNAL'

    #: Tell the server to replace addresses on future SOCKS requests.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.map_address`
    #:     - Command implementation: :class:`CommandMapAddress`
    #:     - Reply implementation: :class:`.ReplyMapAddress`
    MAPADDRESS = 'MAPADDRESS'

    #: Get server information.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.get_info`
    #:     - Command implementation: :class:`CommandGetInfo`
    #:     - Reply implementation: :class:`.ReplyGetInfo`
    GETINFO = 'GETINFO'

    #: Build a new or extend an existing circuit.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.extend_circuit`
    #:     - Command implementation: :class:`CommandExtendCircuit`
    #:     - Reply implementation: :class:`.ReplyExtendCircuit`
    EXTENDCIRCUIT = 'EXTENDCIRCUIT'

    #: Change the purpose of a circuit.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.set_circuit_purpose`
    #:     - Command implementation: :class:`CommandSetCircuitPurpose`
    #:     - Reply implementation: :class:`.ReplySetCircuitPurpose`
    SETCIRCUITPURPOSE = 'SETCIRCUITPURPOSE'

    #: Not implemented because it was marked as obsolete as of ``Tor v0.2.0.8``.
    SETROUTERPURPOSE = 'SETROUTERPURPOSE'

    #: Request that the specified stream should be associated with the specified circuit.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.attach_stream`
    #:     - Command implementation: :class:`CommandAttachStream`
    #:     - Reply implementation: :class:`.ReplyAttachStream`
    ATTACHSTREAM = 'ATTACHSTREAM'

    #: This message informs the server about a new router descriptor.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.post_descriptor`
    #:     - Command implementation: :class:`CommandPostDescriptor`
    #:     - Reply implementation: :class:`.ReplyPostDescriptor`
    POSTDESCRIPTOR = 'POSTDESCRIPTOR'

    #: Tells the server to change the exit address on the specified stream.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.redirect_stream`
    #:     - Command implementation: :class:`CommandRedirectStream`
    #:     - Reply implementation: :class:`.ReplyRedirectStream`
    REDIRECTSTREAM = 'REDIRECTSTREAM'

    #: Tells the server to close the specified stream.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.close_stream`
    #:     - Command implementation: :class:`CommandCloseStream`
    #:     - Reply implementation: :class:`.ReplyCloseStream`
    CLOSESTREAM = 'CLOSESTREAM'

    #: Tells the server to close the specified circuit.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.close_circuit`
    #:     - Command implementation: :class:`CommandCloseCircuit`
    #:     - Reply implementation: :class:`.ReplyCloseCircuit`
    CLOSECIRCUIT = 'CLOSECIRCUIT'

    #: Tells the server to hang up on this controller connection.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.quit`
    #:     - Command implementation: :class:`CommandQuit`
    #:     - Reply implementation: :class:`.ReplyQuit`
    QUIT = 'QUIT'

    #: Enable additional features.
    #:
    #: See Also:
    #:     - Command implementation: :class:`CommandUseFeature`
    #:     - Reply implementation: :class:`.ReplyUseFeature`
    USEFEATURE = 'USEFEATURE'

    #: This command launches a remote hostname lookup request for every specified request.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.resolve`
    #:     - Command implementation: :class:`CommandResolve`
    #:     - Reply implementation: :class:`.ReplyResolve`
    RESOLVE = 'RESOLVE'

    #: This command tells the controller what kinds of authentication are supported.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.protocol_info`
    #:     - Command implementation: :class:`CommandProtocolInfo`
    #:     - Reply implementation: :class:`.ReplyProtocolInfo`
    PROTOCOLINFO = 'PROTOCOLINFO'

    #: This command allows to upload the text of a config file to Tor over the control port.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.load_conf`
    #:     - Command implementation: :class:`CommandLoadConf`
    #:     - Reply implementation: :class:`.ReplyLoadConf`
    LOADCONF = 'LOADCONF'

    #: Instructs Tor to shut down when this control connection is closed.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.take_ownership`
    #:     - Command implementation: :class:`CommandTakeOwnership`
    #:     - Reply implementation: :class:`.ReplyTakeOwnership`
    TAKEOWNERSHIP = 'TAKEOWNERSHIP'

    #: Begin the authentication routine for the SAFECOOKIE method.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.auth_challenge`
    #:     - Command implementation: :class:`CommandAuthChallenge`
    #:     - Reply implementation: :class:`.ReplyAuthChallenge`
    AUTHCHALLENGE = 'AUTHCHALLENGE'

    #: Tells the server to drop all guard nodes.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.drop_guards`
    #:     - Command implementation: :class:`CommandDropGuards`
    #:     - Reply implementation: :class:`.ReplyDropGuards`
    DROPGUARDS = 'DROPGUARDS'

    #: Launches hidden service descriptor fetch(es).
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.hs_fetch`
    #:     - Command implementation: :class:`CommandHsFetch`
    #:     - Reply implementation: :class:`.ReplyHsFetch`
    HSFETCH = 'HSFETCH'

    #: Tells the server to create a new onion "hidden" service.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.add_onion`
    #:     - Command implementation: :class:`CommandAddOnion`
    #:     - Reply implementation: :class:`.ReplyAddOnion`
    ADD_ONION = 'ADD_ONION'

    #: Tells the server to remove an onion "hidden" service.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.del_onion`
    #:     - Command implementation: :class:`CommandDelOnion`
    #:     - Reply implementation: :class:`.ReplyDelOnion`
    DEL_ONION = 'DEL_ONION'

    #: This command launches a hidden service descriptor upload to the specified HSDirs.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.hs_post`
    #:     - Command implementation: :class:`CommandHsPost`
    #:     - Reply implementation: :class:`.ReplyHsPost`
    HSPOST = 'HSPOST'

    #: Add client-side v3 client auth credentials for a onion service.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.onion_client_auth_add`
    #:     - Command implementation: :class:`CommandOnionClientAuthAdd`
    #:     - Reply implementation: :class:`.ReplyOnionClientAuthAdd`
    ONION_CLIENT_AUTH_ADD = 'ONION_CLIENT_AUTH_ADD'

    #: Remove client-side v3 client auth credentials for a onion service.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.onion_client_auth_remove`
    #:     - Command implementation: :class:`CommandOnionClientAuthRemove`
    #:     - Reply implementation: :class:`.ReplyOnionClientAuthRemove`
    ONION_CLIENT_AUTH_REMOVE = 'ONION_CLIENT_AUTH_REMOVE'

    #: List client-side v3 client auth credentials for a onion service.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.onion_client_auth_view`
    #:     - Command implementation: :class:`CommandOnionClientAuthView`
    #:     - Reply implementation: :class:`.ReplyOnionClientAuthView`
    ONION_CLIENT_AUTH_VIEW = 'ONION_CLIENT_AUTH_VIEW'

    #: This command instructs Tor to relinquish ownership of its control connection.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.drop_ownership`
    #:     - Command implementation: :class:`CommandDropOwnership`
    #:     - Reply implementation: :class:`.ReplyDropOwnership`
    DROPOWNERSHIP = 'DROPOWNERSHIP'

    #: Tells the server to drop all circuit build times.
    #:
    #: See Also:
    #:     - Controller method: :meth:`.Controller.drop_timeouts`
    #:     - Command implementation: :class:`CommandDropTimeouts`
    #:     - Reply implementation: :class:`.ReplyDropTimeouts`
    DROPTIMEOUTS = 'DROPTIMEOUTS'


class CommandSerializer:
    """Helper class used to serialize an existing command."""

    #: End of line to use while serializing a command.
    END_OF_LINE: ClassVar[str] = '\r\n'

    def __init__(self, name: CommandWord) -> None:
        """
        Create a new command serializer.

        This is used internally by :meth:`.Command.serialize`.

        Args:
            name: The command name.

        """
        self._body = None  # type: str | None
        self._command = name
        self._arguments = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

    def serialize(self) -> str:
        """
        Serialize the arguments to a string.

        Returns:
            Text that can be pushed to the server.

        """
        # Build the header line.
        args = [self._command.value]
        for argument in self._arguments:
            args.append(str(argument))

        header = ' '.join(args)
        # Check for command injection in case some user-controlled values went through.
        if any(c in header for c in '\r\v\n'):
            msg = 'Command injection was detected and prevented'
            raise CommandError(msg)
        lines = [header]

        # Include the potential body, if applicable.
        if self._body is None:
            prefix = ''
        else:
            for line in self._body.splitlines():
                if line.startswith('.'):
                    line = '.' + line
                lines.append(line)
            lines.append('.')
            prefix = '+'
        return prefix + self.END_OF_LINE.join(lines) + self.END_OF_LINE

    @property
    def command(self) -> CommandWord:
        """Get the command name for the underlying command."""
        return self._command

    @property
    def arguments(self) -> MutableSequence[ArgumentKeyword | ArgumentString]:
        """Get the list of command arguments."""
        return self._arguments

    @property
    def body(self) -> str | None:
        """Get the command body, if any."""
        return self._body

    @body.setter
    def body(self, body: str) -> None:
        """
        Set the command body.

        Args:
            body: The new body content for the command.

        """
        self._body = body


class Command:
    """Base class for all commands."""

    #: Cached adapter used while serializing the command.
    ADAPTER: ClassVar[TypeAdapter[Self] | None] = None

    #: Command word this command is for.
    command: ClassVar[CommandWord]

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        """Create a pydantic validator and serializer for this structure."""
        # Is there another way to simply wrap an existing schema?
        # We simply need to add a custom serializer on top of the existing one.
        return core_schema.union_schema(
            choices=[handler(source)],
            serialization=core_schema.wrap_serializer_function_ser_schema(
                function=cls._pydantic_serializer,
                schema=handler(source),
                return_schema=core_schema.str_schema(),
            ),
        )

    @classmethod
    def _pydantic_serializer(cls, item: Self, to_dict: SerializerFunctionWrapHandler) -> str:
        """
        Serialize the provided command to a nice string.

        This method is the one used by pydantic during serialization.
        Here ``to_dict`` is a function used to call the inner (original) serializer,
        which provides a nice dictionary with all of our fields already serializer.

        """
        return item.serialize_from_struct(to_dict(item))

    @classmethod
    def adapter(cls) -> TypeAdapter[Self]:
        """Get a cached type adapter to serialize a command."""
        if cls.ADAPTER is None:
            cls.ADAPTER = TypeAdapter(cls)
        return cls.ADAPTER

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """
        Serialize the command to text.

        This command is not intended to be called by the end user.

        Args:
            struct: a dictionary serialization for this structure.

        Returns:
            Text that can be sent to Tor's control port.

        """
        return CommandSerializer(self.command).serialize()

    def serialize(self) -> str:
        """Serialize this command to a string."""
        return self.adapter().dump_python(self)


@dataclass(kw_only=True)
class CommandSetConf(Command):
    """
    Command implementation for :attr:`~CommandWord.SETCONF`.

    Change the value of one or more configuration variables.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#setconf

    """

    command: ClassVar[CommandWord] = CommandWord.SETCONF

    #: All the configuration values you want to set.
    values: MutableMapping[str, MutableSequence[int | str] | int | str | None] = field(
        default_factory=dict,
    )

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``SETCONF`` specific arguments."""
        values = struct['values']
        if len(values) == 0:
            msg = f"No value provided for command '{self.command.value}'"
            raise CommandError(msg)

        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        for key, value in values.items():
            if isinstance(value, MutableSequence):
                for item in value:
                    args.append(ArgumentKeyword(key, item))
            else:
                args.append(ArgumentKeyword(key, value))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandResetConf(CommandSetConf):
    """
    Command implementation for :attr:`~CommandWord.RESETCONF`.

    Remove all settings for a given configuration option entirely,
    assign its default value (if any), and then assign the value provided.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#resetconf

    """

    command: ClassVar[CommandWord] = CommandWord.RESETCONF


@dataclass(kw_only=True)
class CommandGetConf(Command):
    """
    Command implementation for :attr:`~CommandWord.GETCONF`.

    Request the value of zero or more configuration variable(s).

    See Also:
        https://spec.torproject.org/control-spec/commands.html#getconf

    """

    command: ClassVar[CommandWord] = CommandWord.GETCONF

    #: List of configuration keys to request (duplicates mean duplicate answers).
    keywords: MutableSequence[str] = field(default_factory=list)

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``GETCONF`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        for keyword in struct['keywords']:
            args.append(ArgumentString(keyword))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandSetEvents(Command):
    """
    Command implementation for :attr:`~CommandWord.SETEVENTS`.

    Request the server to inform the client about interesting events.

    See Also:
        - https://spec.torproject.org/control-spec/commands.html#setevents
        - :meth:`.Controller.add_event_handler`

    """

    command: ClassVar[CommandWord] = CommandWord.SETEVENTS

    #: Set of event names to receive the corresponding events.
    events: set[EventWord] = field(default_factory=set)

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``SETEVENTS`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        for evt in struct['events']:
            args.append(ArgumentString(evt, safe=True))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandAuthenticate(Command):
    """
    Command implementation for :attr:`~CommandWord.AUTHENTICATE`.

    This command is used to authenticate to the server.

    See Also:
        - https://spec.torproject.org/control-spec/commands.html#authenticate
        - :meth:`.Controller.authenticate` and :attr:`.Controller.authenticated`.

    """

    command: ClassVar[CommandWord] = CommandWord.AUTHENTICATE

    #: Password or token used to authenticate with the server.
    token: Base16Bytes | str | None = None

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``AUTHENTICATE`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        # Here we need to have the original type for serialization.
        token = struct['token']
        match self.token:
            case bytes():
                args.append(ArgumentKeyword(None, token, quotes=QuoteStyle.NEVER))
            case str():
                args.append(ArgumentKeyword(None, token, quotes=QuoteStyle.ALWAYS))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandSaveConf(Command):
    """
    Command implementation for :attr:`~CommandWord.SAVECONF`.

    Instructs the server to write out its configuration options into ``torrc``.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#saveconf

    """

    command: ClassVar[CommandWord] = CommandWord.SAVECONF

    #: If ``%include`` is used on ``torrc``, ``SAVECONF`` will not write the configuration
    #: to disk.  When set, the configuration will be overwritten even if %include is used.
    #: You can find out whether this flag is needed using ``config-can-saveconf`` on
    #: :class:`CommandGetInfo`.
    force: bool = False

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``SAVECONF`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        if struct['force'] is True:
            # Flags are treated as keywords with no value.
            args.append(ArgumentKeyword('FORCE', None))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandSignal(Command):
    """
    Command implementation for :attr:`~CommandWord.SIGNAL`.

    Send a signal to Tor.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#signal

    """

    command: ClassVar[CommandWord] = CommandWord.SIGNAL

    #: The signal to send to Tor.
    signal: Signal

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``SIGNAL`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        args.append(ArgumentString(struct['signal'], safe=True))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandMapAddress(Command):
    """
    Command implementation for :attr:`~CommandWord.MAPADDRESS`.

    The client sends this message to the server in order to tell it that future
    SOCKS requests for connections to the original address should be replaced
    with connections to the specified replacement address.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#mapaddress

    """

    command: ClassVar[CommandWord] = CommandWord.MAPADDRESS

    #: Map of addresses to remap on socks requests.
    addresses: MutableMapping[Union[AnyHost], Union[AnyHost]] = field(default_factory=dict)  # noqa: UP007

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``MAPADDRESS`` specific arguments."""
        addresses = struct['addresses']
        if len(addresses) == 0:
            msg = "No address provided for command 'MAPADDRESS'"
            raise CommandError(msg)

        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        for key, value in addresses.items():
            args.append(ArgumentKeyword(key, value, quotes=QuoteStyle.NEVER_ENSURE))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandGetInfo(Command):
    """
    Command implementation for :attr:`~CommandWord.GETINFO`.

    Unlike :attr:`~CommandWord.GETCONF` this message is used for data that are not stored
    in the Tor configuration file, and that may be longer than a single line.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#getinfo

    """

    command: ClassVar[CommandWord] = CommandWord.GETINFO

    #: List of keywords to request the value from. One or more must be provided.
    keywords: MutableSequence[str] = field(default_factory=list)

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``GETINFO`` specific arguments."""
        keywords = struct['keywords']
        if len(keywords) == 0:
            msg = "No keyword provided for command 'GETINFO'"
            raise CommandError(msg)

        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        for keyword in keywords:
            args.append(ArgumentString(keyword))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandExtendCircuit(Command):
    """
    Command implementation for :attr:`~CommandWord.EXTENDCIRCUIT`.

    This request takes one of two forms: either :attr:`circuit` is zero, in which case it is
    a request for the server to build a new circuit, or :attr:`circuit` is nonzero, in which
    case it is a request for the server to extend an existing circuit with that ID
    according to the specified path provided in :attr:`servers`.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#extendcircuit

    """

    command: ClassVar[CommandWord] = CommandWord.EXTENDCIRCUIT

    #: Circuit identifier to extend, ``0`` to create a new circuit.
    circuit: int

    #: List of servers to extend the circuit onto (or no server).
    servers: Annotated[
        MutableSequence[LongServerName],
        TrBeforeStringSplit(),
    ] = field(default_factory=list)

    #: Circuit purpose or :obj:`None` to use a default purpose.
    purpose: Literal[CircuitPurpose.CONTROLLER, CircuitPurpose.GENERAL] | None = None

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``EXTENDCIRCUIT`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        circuit = struct['circuit']
        args.append(ArgumentString(circuit, safe=True))

        servers = struct['servers']
        if servers:
            args.append(ArgumentString(servers))

        purpose = struct['purpose']
        if purpose is not None:
            args.append(ArgumentKeyword('purpose', purpose, quotes=QuoteStyle.NEVER))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandSetCircuitPurpose(Command):
    """
    Command implementation for :attr:`~CommandWord.SETCIRCUITPURPOSE`.

    This changes the descriptor's purpose.

    Hints:
        See :class:`CommandPostDescriptor` for more details on :attr:`purpose`.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#setcircuitpurpose

    """

    command: ClassVar[CommandWord] = CommandWord.SETCIRCUITPURPOSE

    #: Circuit ID to set the purpose on.
    circuit: int

    #: Set purpose of the provided circuit.
    purpose: Literal[CircuitPurpose.CONTROLLER, CircuitPurpose.GENERAL]

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``SETCIRCUITPURPOSE`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        args.append(ArgumentString(struct['circuit'], safe=True))
        args.append(ArgumentKeyword('purpose', struct['purpose'], quotes=QuoteStyle.NEVER))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandAttachStream(Command):
    """
    Command implementation for :attr:`~CommandWord.ATTACHSTREAM`.

    This message informs the server that the specified :attr:`stream` should
    be associated with the :attr:`circuit`.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#attachstream

    """

    command: ClassVar[CommandWord] = CommandWord.ATTACHSTREAM

    #: Stream to associate to the provided circuit.
    stream: int

    #: Circuit identifier to attach the stream onto.
    circuit: int

    #: When set, Tor will choose the HopNumth hop in the circuit as the exit node,
    #: rather that the last node in the circuit. Hops are 1-indexed; generally,
    #: it is not permitted to attach to hop 1.
    hop: int | None = None

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``ATTACHSTREAM`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        args.append(ArgumentString(struct['stream'], safe=True))
        args.append(ArgumentString(struct['circuit'], safe=True))

        hop = struct['hop']
        if hop is not None:
            args.append(ArgumentKeyword('HOP', hop, quotes=QuoteStyle.NEVER))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandPostDescriptor(Command):
    """
    Command implementation for :attr:`~CommandWord.POSTDESCRIPTOR`.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#postdescriptor

    """

    command: ClassVar[CommandWord] = CommandWord.POSTDESCRIPTOR

    #: Purpose of the provided descriptor.
    purpose: DescriptorPurpose | None = None

    #: Cache the provided descriptor internally.
    cache: BoolYesNo | None = None

    #: Descriptor content.
    descriptor: str

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``POSTDESCRIPTOR`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        purpose = struct['purpose']
        if purpose is not None:
            args.append(ArgumentKeyword('purpose', purpose, quotes=QuoteStyle.NEVER))

        cache = struct['cache']
        if cache is not None:
            args.append(ArgumentKeyword('cache', cache, quotes=QuoteStyle.NEVER))

        ser.arguments.extend(args)
        ser.body = struct['descriptor']
        return ser.serialize()


@dataclass(kw_only=True)
class CommandRedirectStream(Command):
    """
    Command implementation for :attr:`~CommandWord.REDIRECTSTREAM`.

    Tells the server to change the exit address on the specified stream.
    If :attr:`port` is specified, changes the destination port as well.

    No remapping is performed on the new provided address.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#redirectstream

    """

    command: ClassVar[CommandWord] = CommandWord.REDIRECTSTREAM

    #: Stream identifier to redirect.
    stream: int
    #: Destination address to redirect it to.
    address: Union[AnyHost]  # noqa: UP007
    #: Optional port to redirect the stream to.
    port: AnyPort | None = None

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``REDIRECTSTREAM`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        args.append(ArgumentString(struct['stream'], safe=True))
        args.append(ArgumentString(struct['address']))

        port = struct['port']
        if port is not None:
            args.append(ArgumentString(port, safe=True))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandCloseStream(Command):
    """
    Command implementation for :attr:`~CommandWord.CLOSESTREAM`.

    Tells the server to close the specified :attr:`stream`.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#closestream

    """

    command: ClassVar[CommandWord] = CommandWord.CLOSESTREAM

    #: Identifier to the stream to close.
    stream: int
    #: Provide a reason for the stream to be closed.
    reason: StreamCloseReasonInt

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``CLOSESTREAM`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        args.append(ArgumentString(struct['stream'], safe=True))
        args.append(ArgumentString(struct['reason'], safe=True))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandCloseCircuit(Command):
    """
    Command implementation for :attr:`~CommandWord.CLOSECIRCUIT`.

    Tells the server to close the specified circuit.

    When :attr:`if_unused` is :obj:`True`, do not close the circuit unless it is unused.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#closecircuit

    """

    command: ClassVar[CommandWord] = CommandWord.CLOSECIRCUIT

    #: Circuit identifier to close.
    circuit: int

    #: Do not close the circuit unless it is unused.
    if_unused: bool = False

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``CLOSECIRCUIT`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        args.append(ArgumentString(struct['circuit'], safe=True))

        if struct['if_unused'] is True:
            args.append(ArgumentKeyword('IfUnused', None))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandQuit(Command):
    """
    Command implementation for :attr:`~CommandWord.QUIT`.

    Tells the server to hang up on this controller connection.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#quit

    """

    command: ClassVar[CommandWord] = CommandWord.QUIT


@dataclass(kw_only=True)
class CommandUseFeature(Command):
    """
    Command implementation for :attr:`~CommandWord.USEFEATURE`.

    Adding additional features to the control protocol sometimes will break backwards
    compatibility. Initially such features are added into Tor and disabled by default.
    :attr:`~CommandWord.USEFEATURE` can enable these additional features.

    Note:
        To get a list of available features please use ``features/names``
        with :class:`CommandGetInfo`.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#usefeature

    """

    command: ClassVar[CommandWord] = CommandWord.USEFEATURE

    #: Set of features to enable.
    features: set[Feature | str] = field(default_factory=set)

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``USEFEATURE`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        for feature in struct['features']:
            args.append(ArgumentString(feature))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandResolve(Command):
    """
    Command implementation for :attr:`~CommandWord.RESOLVE`.

    This command launches a remote hostname lookup request for every specified
    request (or reverse lookup if :attr:`reverse` is specified).
    Note that the request is done in the background: to see the answers, your controller
    will need to listen for :attr:`.EventWord.ADDRMAP` events.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#resolve

    """

    command: ClassVar[CommandWord] = CommandWord.RESOLVE

    #: List of addresses get a resolution for.
    addresses: MutableSequence[Union[AnyHost]] = field(default_factory=list)  # noqa: UP007
    #: Whether we should perform a reverse lookup resolution.
    reverse: bool = False

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``RESOLVE`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        if struct['reverse'] is True:
            args.append(ArgumentKeyword('mode', 'reverse', quotes=QuoteStyle.NEVER))
        for address in struct['addresses']:
            # These are marked as keywords in `src/feature/control/control_cmd.c`.
            args.append(ArgumentKeyword(address, None))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandProtocolInfo(Command):
    """
    Command implementation for :attr:`~CommandWord.PROTOCOLINFO`.

    This command tells the controller what kinds of authentication are supported.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#protocolinfo

    """

    command: ClassVar[CommandWord] = CommandWord.PROTOCOLINFO

    #: Optional version to request information for (ignored by Tor at the moment).
    version: int | None = None

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``PROTOCOLINFO`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        version = struct['version']
        if version is not None:
            args.append(ArgumentString(version, safe=True))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandLoadConf(Command):
    """
    Command implementation for :attr:`~CommandWord.LOADCONF`.

    This command allows a controller to upload the text of a config file to Tor over
    the control port. This config file is then loaded as if it had been read from disk.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#loadconf

    """

    command: ClassVar[CommandWord] = CommandWord.LOADCONF

    #: Raw configuration text to load.
    text: str

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``LOADCONF`` specific arguments."""
        ser = CommandSerializer(self.command)
        ser.body = struct['text']
        return ser.serialize()


@dataclass(kw_only=True)
class CommandTakeOwnership(Command):
    """
    Command implementation for :attr:`~CommandWord.TAKEOWNERSHIP`.

    This command instructs Tor to shut down when this control connection is closed.
    It affects each control connection that sends it independently; if multiple control
    connections send the :attr:`~CommandWord.TAKEOWNERSHIP` command to a Tor instance,
    Tor will shut down when any of those connections closes.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#takeownership

    """

    command: ClassVar[CommandWord] = CommandWord.TAKEOWNERSHIP


@dataclass(kw_only=True)
class CommandAuthChallenge(Command):
    """
    Command implementation for :attr:`~CommandWord.AUTHCHALLENGE`.

    This command is used to begin the authentication routine for the
    :attr:`~.AuthMethod.SAFECOOKIE` authentication method.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#authchallenge

    """

    #: Length of the nonce we expect to receive (when :class:`bytes`).
    NONCE_LENGTH: ClassVar[int] = 32

    command: ClassVar[CommandWord] = CommandWord.AUTHCHALLENGE

    #: Nonce value, a new one is generated when none is provided.
    nonce: Base16Bytes | str

    @classmethod
    def generate_nonce(cls) -> bytes:
        """Generate a nonce value of 32 bytes."""
        return secrets.token_bytes(cls.NONCE_LENGTH)

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``AUTHCHALLENGE`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        args.append(ArgumentString('SAFECOOKIE', safe=True))

        # Here we need to have the original type for serialization.
        nonce = struct['nonce']
        match self.nonce:
            case bytes():
                args.append(ArgumentKeyword(None, nonce, quotes=QuoteStyle.NEVER))
            case str():  # pragma: no branch
                args.append(ArgumentKeyword(None, nonce, quotes=QuoteStyle.ALWAYS))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandDropGuards(Command):
    """
    Command implementation for :attr:`~CommandWord.DROPGUARDS`.

    Tells the server to drop all guard nodes.

    Warning:
        Do not invoke this command lightly; it can increase vulnerability to
        tracking attacks over time.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#dropguards

    """

    command: ClassVar[CommandWord] = CommandWord.DROPGUARDS


@dataclass(kw_only=True)
class CommandHsFetch(Command):
    """
    Command implementation for :attr:`~CommandWord.HSFETCH`.

    This command launches hidden service descriptor fetch(es) for the given :attr:`address`.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#hsfetch

    """

    command: ClassVar[CommandWord] = CommandWord.HSFETCH

    #: Optional list of servers to contact for a hidden service descriptor.
    servers: MutableSequence[LongServerName] = field(default_factory=list)

    #: Onion address (v2 or v3) to request a descriptor for, without the ``.onion`` suffix.
    address: HiddenServiceAddress

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``HSFETCH`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        args.append(ArgumentString(struct['address'], safe=True))
        for server in struct['servers']:
            args.append(ArgumentKeyword('SERVER', server, quotes=QuoteStyle.NEVER_ENSURE))
        ser.arguments.extend(args)
        return ser.serialize()


def _onion_add_get_key_tag(v: Any) -> str:
    """Determine which kind of key we are using (existing or new)."""
    match v:
        case str():
            if v.startswith('NEW:'):
                return 'new'
        case OnionServiceNewKeyStruct():
            return 'new'
    return 'key'


@dataclass(kw_only=True)
class CommandAddOnion(Command):
    """
    Command implementation for :attr:`~CommandWord.ADD_ONION`.

    Tells Tor to create a new onion "hidden" Service, with the specified private key
    and algorithm.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#add_onion

    """

    command: ClassVar[CommandWord] = CommandWord.ADD_ONION

    #: The service key, either provided directly of generated by Tor.
    key: Annotated[
        Union[  # noqa: UP007
            Annotated[OnionServiceKey, Tag('key')],
            Annotated[OnionServiceNewKey, Tag('new')],
        ],
        Discriminator(_onion_add_get_key_tag),
    ]

    #: Set of boolean options to attach to this service.
    flags: Annotated[set[OnionServiceFlags], TrBeforeStringSplit()] = field(
        default_factory=set
    )

    #: Optional number between 0 and 65535 which is the maximum streams that can be
    #: attached on a rendezvous circuit. Setting it to 0 means unlimited which is
    #: also the default behavior.
    max_streams: NonNegativeInt | None = None

    #: As in an arguments to config ``HiddenServicePort``, ``port,target``.
    ports: MutableSequence[VirtualPort] = field(default_factory=list)

    #: Client authentications for Onion V2, syntax is ``ClientName[:ClientBlob]``.
    client_auth: MutableSequence[HsDescClientAuthV2] = field(default_factory=list)

    #: String syntax is a base32-encoded ``x25519`` public key with only the key part.
    client_auth_v3: MutableSequence[HsDescClientAuthV3] = field(default_factory=list)

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``ADD_ONION`` specific arguments."""
        ports = struct['ports']
        if not len(ports):
            msg = 'You must specify one or more virtual ports.'
            raise CommandError(msg)

        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        args.append(ArgumentString(struct['key']))

        flags = struct['flags']
        if len(flags):
            args.append(ArgumentKeyword('Flags', flags, quotes=QuoteStyle.NEVER))

        max_streams = struct['max_streams']
        if max_streams is not None:
            kwarg = ArgumentKeyword('MaxStreams', max_streams, quotes=QuoteStyle.NEVER)
            args.append(kwarg)

        for port in ports:
            args.append(ArgumentKeyword('Port', port, quotes=QuoteStyle.NEVER_ENSURE))

        for auth in struct['client_auth']:
            args.append(ArgumentKeyword('ClientAuth', auth, quotes=QuoteStyle.NEVER_ENSURE))

        for auth in struct['client_auth_v3']:
            args.append(ArgumentKeyword('ClientAuthV3', auth, quotes=QuoteStyle.NEVER))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandDelOnion(Command):
    """
    Command implementation for :attr:`~CommandWord.DEL_ONION`.

    Tells the server to remove an Onion "hidden" Service, that was previously created
    trough :class:`CommandAddOnion`. It is only possible to remove onion services that were
    created on the same control connection as the :attr:`~CommandWord.DEL_ONION` command, and
    those that belong to no control connection in particular
    (the :attr:`~.OnionServiceFlags.DETACH` flag was specified upon creation).

    See Also:
        https://spec.torproject.org/control-spec/commands.html#del_onion

    """

    command: ClassVar[CommandWord] = CommandWord.DEL_ONION

    #: This is the v2 or v3 address without the ``.onion`` suffix.
    address: HiddenServiceAddress

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``DEL_ONION`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        args.append(ArgumentString(struct['address']))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandHsPost(Command):
    """
    Command implementation for :attr:`~CommandWord.HSPOST`.

    This command launches a hidden service descriptor upload to the specified HSDirs.
    If one or more Server arguments are provided, an upload is triggered on each of
    them in parallel. If no Server options are provided, it behaves like a normal HS
    descriptor upload and will upload to the set of responsible HS directories.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#hspost

    """

    command: ClassVar[CommandWord] = CommandWord.HSPOST

    #: List of servers to upload the descriptor to (if any is provided).
    servers: MutableSequence[LongServerName] = field(default_factory=list)
    #: This is the optional v2 or v3 address without the ``.onion`` suffix.
    address: HiddenServiceAddress | None = None
    #: Descriptor content as raw text.
    descriptor: str

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``HSPOST`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        for server in struct['servers']:
            args.append(ArgumentKeyword('SERVER', server, quotes=QuoteStyle.NEVER_ENSURE))

        address = struct['address']
        if address is not None:
            kwarg = ArgumentKeyword('HSADDRESS', address, quotes=QuoteStyle.NEVER_ENSURE)
            args.append(kwarg)

        ser.arguments.extend(args)
        ser.body = struct['descriptor']
        return ser.serialize()


@dataclass(kw_only=True)
class CommandOnionClientAuthAdd(Command):
    """
    Command implementation for :attr:`~CommandWord.ONION_CLIENT_AUTH_ADD`.

    Tells the connected Tor to add client-side v3 client auth credentials for the onion
    service with :attr:`address`. The :attr:`key` is the x25519 private key that should
    be used for this client, and :attr:`nickname` is an optional nickname for the client.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#onion_client_auth_add

    """

    command: ClassVar[CommandWord] = CommandWord.ONION_CLIENT_AUTH_ADD

    #: V3 onion address without the ``.onion`` suffix.
    address: HiddenServiceAddressV3

    #: The private ``x25519`` key used to authenticate to :attr:`address`.
    key: OnionClientAuthKey

    #: An optional nickname for the client.
    nickname: str | None = None

    #: Whether this client's credentials should be stored on the file system.
    flags: Annotated[set[OnionClientAuthFlags], TrBeforeStringSplit()] = field(
        default_factory=set
    )

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``ONION_CLIENT_AUTH_ADD`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]

        args.append(ArgumentString(struct['address']))
        args.append(ArgumentString(struct['key']))

        nickname = struct['nickname']
        if nickname is not None:
            kwarg = ArgumentKeyword('ClientName', nickname, quotes=QuoteStyle.NEVER_ENSURE)
            args.append(kwarg)

        flags = struct['flags']
        if len(flags):
            args.append(ArgumentKeyword('Flags', flags, quotes=QuoteStyle.NEVER))

        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandOnionClientAuthRemove(Command):
    """
    Command implementation for :attr:`~CommandWord.ONION_CLIENT_AUTH_REMOVE`.

    Tells the connected Tor to remove the client-side v3 client auth credentials
    for the onion service with :attr:`address`.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#onion_client_auth_remove

    """

    command: ClassVar[CommandWord] = CommandWord.ONION_CLIENT_AUTH_REMOVE

    #: V3 onion address without the ``.onion`` suffix.
    address: HiddenServiceAddressV3

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``ONION_CLIENT_AUTH_REMOVE`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        args.append(ArgumentString(struct['address']))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandOnionClientAuthView(Command):
    """
    Command implementation for :attr:`~CommandWord.ONION_CLIENT_AUTH_VIEW`.

    Tells the connected Tor to list all the stored client-side v3 client auth credentials
    for :attr:`address`. If no :attr:`address` is provided, list all the stored client-side
    v3 client auth credentials.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#onion_client_auth_view

    """

    command: ClassVar[CommandWord] = CommandWord.ONION_CLIENT_AUTH_VIEW

    #: V3 onion address without the ``.onion`` suffix.
    address: HiddenServiceAddressV3 | None = None

    def serialize_from_struct(self, struct: Mapping[str, Any]) -> str:
        """Append ``ONION_CLIENT_AUTH_VIEW`` specific arguments."""
        ser = CommandSerializer(self.command)
        args = []  # type: MutableSequence[ArgumentKeyword | ArgumentString]
        address = struct['address']
        if address is not None:
            args.append(ArgumentString(address))
        ser.arguments.extend(args)
        return ser.serialize()


@dataclass(kw_only=True)
class CommandDropOwnership(Command):
    """
    Command implementation for :attr:`~CommandWord.DROPOWNERSHIP`.

    This command instructs Tor to relinquish ownership of its control connection.
    As such tor will not shut down when this control connection is closed.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#dropownership

    """

    command: ClassVar[CommandWord] = CommandWord.DROPOWNERSHIP


@dataclass(kw_only=True)
class CommandDropTimeouts(Command):
    """
    Command implementation for :attr:`~CommandWord.DROPTIMEOUTS`.

    Tells the server to drop all circuit build times.

    Warning:
        Do not invoke this command lightly; it can increase vulnerability
        to tracking attacks over time.

    See Also:
        https://spec.torproject.org/control-spec/commands.html#droptimeouts

    """

    command: ClassVar[CommandWord] = CommandWord.DROPTIMEOUTS



from __future__ import annotations

import asyncio
import os
from typing import TYPE_CHECKING

import pytest
import pytest_asyncio

from aiostem import Controller
from aiostem.command import CommandWord
from aiostem.exceptions import ReplyStatusError
from aiostem.utils import Message

if TYPE_CHECKING:
    from collections.abc import (
        Mapping,
        Sequence,
        Set as AbstractSet,
    )

    from aiostem.controller import EventCallbackType
    from aiostem.utils import (
        Command,
        EventWord,
        ReplyAuthChallenge,
        ReplyProtocolInfo,
        ReplySetEvents,
        ReplySignal,
    )


class CustomController(Controller):
    """A controller finely tuned for tests."""

    def __init__(self, connector) -> None:
        super().__init__(connector)
        self.auth_cookie_data = None
        self.auth_cookie_file = None
        self.error_on_set_events = False
        self.enabled_auth_methods = set()
        self.traces = set()
        self.event_signal_active = asyncio.Event()
        self.trace_commands = []
        self.trace_replies = []
        self.trace_signals = []

    @property
    def event_handlers(self) -> Mapping[str, Sequence[EventCallbackType]]:
        """Direct acccess to the event handlers."""
        return self._evt_callbacks

    async def protocol_info(self, version: int | None = None) -> ReplyProtocolInfo:
        """Filter out some auth_methods when asked to."""
        reply = await super().protocol_info(version)
        if reply.data is not None:
            if self.enabled_auth_methods:
                reply.data.auth_methods = self.enabled_auth_methods
            if self.auth_cookie_file is not None:
                reply.data.auth_cookie_file = self.auth_cookie_file
        return reply

    async def auth_challenge(self, nonce: bytes | str | None = None) -> ReplyAuthChallenge:
        """Fix the auth challenge a little bit when needed."""
        reply = await super().auth_challenge(nonce)
        if reply.data is not None and self.auth_cookie_data is not None:
            reply.data.server_hash = reply.data.build_server_hash(self.auth_cookie_data)
        return reply

    async def push_event_message(self, message: Message) -> None:
        """Push a spurious event for test purposes."""
        await self._on_event_received(message)

    async def request(self, command: Command) -> Message:
        if 'command' in self.traces:
            self.trace_commands.append(command)

        if command.command in frozenset(
            {
                CommandWord.DROPGUARDS,
                CommandWord.DROPOWNERSHIP,
                CommandWord.DROPTIMEOUTS,
                CommandWord.TAKEOWNERSHIP,
            }
        ):
            message = Message(status=250, header='OK')
        else:
            message = await super().request(command)

        if 'replies' in self.traces:
            self.trace_replies.append(message)

        return message

    async def set_events(self, events: AbstractSet[EventWord]) -> ReplySetEvents:
        if self.error_on_set_events:
            msg = 'Triggered by PyTest.'
            raise ReplyStatusError(msg, code=500)
        return await super().set_events(events)

    async def signal(self, signal: str) -> ReplySignal:
        result = await super().signal(signal)

        if 'signal' in self.traces:
            self.trace_signals.append(signal)

        match signal:
            case 'ACTIVE':
                self.event_signal_active.set()

        return result


@pytest.fixture
def controller_raw():
    host = os.environ.get('AIOSTEM_HOST', '127.0.0.1')
    port = int(os.environ.get('AIOSTEM_PORT', 9051))
    return CustomController.from_port(host, port)


@pytest_asyncio.fixture()
async def controller_unauth(controller_raw):
    async with controller_raw:
        yield controller_raw


@pytest_asyncio.fixture()
async def controller(controller_unauth):
    password = os.environ.get('AIOSTEM_PASS', 'aiostem')
    await controller_unauth.authenticate(password)
    return controller_unauth



from __future__ import annotations

import logging

import pytest

from aiostem.exceptions import ReplySyntaxError
from aiostem.utils import Message, MessageData
from aiostem.utils.syntax import ReplySyntax, ReplySyntaxFlag


class TestReplySyntax:
    """Checks on our reply parser."""

    def test_positional(self):
        syntax = ReplySyntax(args_min=2, args_map=['severity', 'message'])
        message = Message(status=650, header='NOTICE HelloWorld')
        result = syntax.parse(message)
        assert len(result) == 2
        assert result['severity'] == 'NOTICE'
        assert result['message'] == 'HelloWorld'

    def test_positional_as_string(self):
        syntax = ReplySyntax(args_min=2, args_map=['severity', 'message'])
        result = syntax.parse_string('NOTICE HelloWorld')
        assert len(result) == 2
        assert result['severity'] == 'NOTICE'
        assert result['message'] == 'HelloWorld'

    def test_positional_with_omission(self):
        syntax = ReplySyntax(args_min=2, args_map=[None, 'message'])
        message = Message(status=650, header='NOTICE HelloWorld')
        result = syntax.parse(message)
        assert len(result) == 1
        assert result['message'] == 'HelloWorld'

    def test_positional_with_remain(self):
        text = 'No user activity in a long time: becoming dormant'
        syntax = ReplySyntax(
            args_min=2,
            args_map=['severity', 'message'],
            flags=ReplySyntaxFlag.POS_REMAIN,
        )
        message = Message(status=650, header=f'NOTICE {text}')
        result = syntax.parse(message)
        assert result['message'] == text

    def test_keyword(self):
        syntax = ReplySyntax(
            args_min=1,
            args_map=['positional'],
            kwargs_map={'ControlPort': 'control_port'},
            flags=ReplySyntaxFlag.KW_ENABLE,
        )
        message = Message(status=650, header='TEST ControlPort=0.0.0.0:9051')
        result = syntax.parse(message)
        assert result['control_port'] == '0.0.0.0:9051'
        assert result['positional'] == 'TEST'

    def test_keyword_quoted(self):
        syntax = ReplySyntax(
            kwargs_map={'KEY': 'key'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        )
        message = Message(status=250, header='KEY="He said \\"Hello world\\"."')
        result = syntax.parse(message)
        assert result['key'] == 'He said "Hello world".'

    def test_keyword_omit_keys(self):
        syntax = ReplySyntax(
            kwargs_map={None: 'flags'},
            kwargs_multi={'flags'},
            flags=(
                ReplySyntaxFlag.KW_ENABLE
                | ReplySyntaxFlag.KW_QUOTED
                | ReplySyntaxFlag.KW_OMIT_KEYS
            ),
        )
        # Some flags are quoted here, because why not!
        message = Message(status=250, header='EXTENDED_EVENTS "VERBOSE_NAMES"')
        result = syntax.parse(message)
        flags = result['flags']
        assert len(flags) == 2
        assert flags == ['EXTENDED_EVENTS', 'VERBOSE_NAMES']

    def test_keyword_omit_value(self):
        syntax = ReplySyntax(
            kwargs_map={
                'EXTENDED_EVENTS': 'EXTENDED_EVENTS',
                'VERBOSE_NAMES': 'VERBOSE_NAMES',
            },
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_OMIT_VALS,
        )
        message = Message(status=250, header='EXTENDED_EVENTS VERBOSE_NAMES')
        result = syntax.parse(message)
        assert set(result.keys()) == set(syntax.kwargs_map.keys())

    def test_keyword_allow_all(self):
        syntax = ReplySyntax(
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_EXTRA,
        )
        message = Message(status=250, header='Server=127.0.0.1 Port=9051')
        result = syntax.parse(message)
        assert len(result) == 2
        assert result['Server'] == '127.0.0.1'
        assert result['Port'] == '9051'

    def test_keyword_ignored(self, caplog):
        syntax = ReplySyntax(
            kwargs_map={'Server': 'Server'},
            flags=ReplySyntaxFlag.KW_ENABLE,
        )
        message = Message(status=250, header='Server=127.0.0.1 Port=9051')
        with caplog.at_level(logging.INFO, logger='aiostem.utils'):
            result = syntax.parse(message)
        assert len(result) == 1
        assert 'Found an unhandled keyword: Port=9051' in caplog.text

    def test_keyword_value_empty_value(self):
        syntax = ReplySyntax(
            kwargs_map={'KEY': 'key'},
            flags=ReplySyntaxFlag.KW_ENABLE,
        )
        message = Message(status=250, header='KEY=')
        result = syntax.parse(message)
        assert result['key'] == ''

    def test_keyword_value_in_data(self):
        syntax = ReplySyntax(
            kwargs_map={'KEY': 'key'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_USE_DATA,
        )
        message = MessageData(status=250, header='KEY=', data='Our value is "here"!')
        result = syntax.parse(message)
        assert result['key'] == message.data

    def test_keyword_value_is_raw(self):
        syntax = ReplySyntax(
            kwargs_map={'KEY': 'key'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_RAW,
        )
        message = MessageData(status=250, header='KEY=A long weird string')
        result = syntax.parse(message)
        assert result['key'] == 'A long weird string'

    def test_bad_parse_too_few_arguments(self):
        syntax = ReplySyntax(args_min=2, args_map=['severity', 'message'])
        message = Message(status=650, header='NOTICE')
        with pytest.raises(ReplySyntaxError, match='Received too few arguments'):
            syntax.parse(message)

    def test_bad_parse_remaining_data(self):
        syntax = ReplySyntax(args_min=2, args_map=['severity', 'message'])
        message = Message(status=650, header='NOTICE Hello world')
        with pytest.raises(ReplySyntaxError, match='Unexpectedly found remaining data:'):
            syntax.parse(message)

    def test_bad_parse_keyword_quote_syntax(self):
        syntax = ReplySyntax(
            kwargs_map={'KEY': 'key'},
            flags=ReplySyntaxFlag.KW_ENABLE | ReplySyntaxFlag.KW_QUOTED,
        )
        message = Message(status=250, header='KEY="Hello word')
        with pytest.raises(ReplySyntaxError, match='No double-quote found before the end'):
            syntax.parse(message)

    def test_bad_parse_keyword_unexpected_quote(self):
        syntax = ReplySyntax(
            kwargs_map={'KEY': 'key'},
            flags=ReplySyntaxFlag.KW_ENABLE,
        )
        message = Message(status=250, header='KEY="Hello word"')
        with pytest.raises(ReplySyntaxError, match='Got an unexpected quoted value'):
            syntax.parse(message)

    def test_bad_parse_no_omit_vals(self):
        syntax = ReplySyntax(
            kwargs_map={
                'EXTENDED_EVENTS': 'EXTENDED_EVENTS',
                'VERBOSE_NAMES': 'VERBOSE_NAMES',
            },
            flags=ReplySyntaxFlag.KW_ENABLE,
        )
        message = Message(status=250, header='EXTENDED_EVENTS VERBOSE_NAMES')
        with pytest.raises(ReplySyntaxError, match='Got a single string without either'):
            syntax.parse(message)

    def test_bad_syntax_min_max(self):
        with pytest.raises(RuntimeError, match='Minimum argument count is greater'):
            ReplySyntax(
                args_min=2,
                args_map=['version'],
            )

    def test_bad_syntax_opt_arg_with_kwarg(self):
        msg = 'Cannot have optional argument along with keyword arguments'
        with pytest.raises(RuntimeError, match=msg):
            ReplySyntax(
                args_min=0,
                args_map=['version'],
                flags=ReplySyntaxFlag.KW_ENABLE,
            )

    def test_bad_syntax_remain_vs_kw(self):
        msg = 'Positional remain and keywords are mutually exclusive'
        with pytest.raises(RuntimeError, match=msg):
            ReplySyntax(flags=ReplySyntaxFlag.POS_REMAIN | ReplySyntaxFlag.KW_ENABLE)

    def test_bad_syntax_keys_vs_vals(self):
        msg = 'KW_OMIT_KEYS and KW_OMIT_VALS are mutually exclusive'
        with pytest.raises(RuntimeError, match=msg):
            ReplySyntax(flags=ReplySyntaxFlag.KW_OMIT_KEYS | ReplySyntaxFlag.KW_OMIT_VALS)

    def test_bad_syntax_raw_vs_quoted(self):
        with pytest.raises(RuntimeError, match='KW_RAW and KW_QUOTED are mutually exclusive'):
            ReplySyntax(flags=ReplySyntaxFlag.KW_RAW | ReplySyntaxFlag.KW_QUOTED)

    def test_bad_syntax_kw_disabled_but_with_kvmap(self):
        with pytest.raises(RuntimeError, match='Keywords are disabled but we found items'):
            ReplySyntax(kwargs_map={'SERVER', 'server'})



from __future__ import annotations

import pytest

from aiostem.command import CommandWord
from aiostem.exceptions import CommandError
from aiostem.utils import ArgumentKeyword, ArgumentString, QuoteStyle


class TestArgument:
    """Check all kind of arguments."""

    @pytest.mark.parametrize(
        ('value', 'string'),
        [
            ('TEST', 'TEST'),
            (123456, '123456'),
            (QuoteStyle.NEVER, '0'),
            (CommandWord.SETCONF, 'SETCONF'),
        ],
    )
    def test_string(self, value, string):
        arg = ArgumentString(value)
        assert arg.value == string
        assert str(arg) == string

    def test_string_error_on_space(self):
        with pytest.raises(CommandError, match='Invalid space in argument'):
            ArgumentString('Hello world')

    def test_string_error_on_invalid_type(self):
        with pytest.raises(CommandError, match='Type object cannot be serialized'):
            ArgumentString(object())

    def test_string_error_on_none(self):
        with pytest.raises(CommandError, match='Value cannot be None'):
            ArgumentString(None)

    @pytest.mark.parametrize(
        ('quotes', 'original', 'escaped'),
        [
            (QuoteStyle.ALWAYS, 'AIOSTEM', '"AIOSTEM"'),
            (QuoteStyle.NEVER, 'AIOSTEM', 'AIOSTEM'),
            (QuoteStyle.AUTO, 'AIOSTEM', 'AIOSTEM'),
            (QuoteStyle.AUTO, 'AIO"STEM', '"AIO\\"STEM"'),
            (QuoteStyle.AUTO, 'AIO\\STEM', '"AIO\\\\STEM"'),
        ],
    )
    def test_keyword(self, quotes, original, escaped):
        """Check keyword argument methods and properties."""
        arg = ArgumentKeyword('key', original, quotes=quotes)
        assert arg.key == 'key'
        assert arg.value == original
        assert arg.quotes == quotes
        assert str(arg) == f'key={escaped}'

    def test_keyword_with_none_key(self):
        """Check that a key of None only returns the value."""
        arg = ArgumentKeyword(None, 'value')
        assert arg.key is None
        assert str(arg) == 'value'

    def test_keyword_with_none_value(self):
        """
        Check that a value of None only returns the key.

        This is allowed in special occasions such as `SETCONF` or `RESETCONF`.
        """
        arg = ArgumentKeyword('key', None)
        assert arg.value is None
        assert str(arg) == 'key'

    def test_keyword_with_none_value_ignored_quotes(self):
        """Check that a none value means that quotes are enforced to NEVER."""
        arg = ArgumentKeyword('key', None, quotes=QuoteStyle.ALWAYS)
        assert arg.quotes == QuoteStyle.NEVER

    @pytest.mark.parametrize(
        'original',
        [
            'C:\\windows\\system',
            'This string contains spaces',
            'qu"ote',
        ],
    )
    def test_keyword_quote_error(self, original: str):
        arg = ArgumentKeyword('key', original, quotes=QuoteStyle.NEVER_ENSURE)
        with pytest.raises(CommandError, match='Argument is only safe with quotes'):
            str(arg)

    def test_keyword_key_and_value_none(self):
        with pytest.raises(CommandError, match='Both key and value cannot be None'):
            ArgumentKeyword(None, None)
